"""
Methods about git.
"""
import git
import os
from functools import lru_cache
from typing import Tuple
from git import Repo, Commit

# from lumo.utils.keys import FN, CFG
from lumo.utils.paths import compare_path
from lumo.proc.path import git_dir

bin_file = ['*.pth', '*.npy', '*.ckpt',
            '*.ft',  # for feather
            '*.pkl'
            ]

lib_gitignores = ['.lumo/',
                  '*.lumo.*',
                  '.data',
                  '.datas',
                  '.dataset',
                  '.datasets',
                  ] + bin_file

py_gitignore = "\n".join(['# Byte-compiled / optimized / DLL files',
                          '__pycache__/', '*.py[cod]',
                          '*$py.class',
                          '# C extensions', '*.so',
                          '# Distribution / packaging',
                          '.Python', 'build/', 'develop-eggs/', 'dist/', 'downloads/', 'eggs/', '.eggs/',
                          'lib/', 'lib64/', 'parts/', 'sdist/', 'var/', 'wheels/', 'pip-wheel-metadata/',
                          'share/python-wheels/', '*.egg-info/', '.installed.cfg', '*.egg', 'MANIFEST',
                          '', '# PyInstaller',
                          '#  Usually these files are written by a python script from a template',
                          '#  before PyInstaller builds the exe, so as to inject date/other infos into '
                          'it.',
                          '*.manifest', '*.spec',
                          '# Installer logs', 'pip-log.txt',
                          'pip-delete-this-directory.txt',
                          '# Unit test / coverage reports',
                          'htmlcov/', '.tox/', '.nox/', '.coverage', '.coverage.*', '.cache',
                          'nosetests.xml', 'coverage.xml', '*.cover', '.hypothesis/', '.pytest_cache/',
                          '# Translations', '*.mo', '*.pot',
                          '# Django stuff:', '*.log', 'local_settings.py', 'db.sqlite3',
                          '# Flask stuff:',
                          'instance/', '.webassets-cache', '', '# Scrapy stuff:', '.scrapy', '',
                          '# Sphinx documentation', 'docs/_build/', '', '# PyBuilder', 'target/', '',
                          '# Jupyter Notebook', '.ipynb_checkpoints', '',
                          '# IPython',
                          'profile_default/', 'ipython_config.py',
                          '# pyenv', '.python-version', '',
                          '# celery beat schedule file', 'celerybeat-schedule', '',
                          '# SageMath parsed files', '*.sage.py', '',
                          '# Environments', '.env', '.venv',
                          'env/', 'venv/', 'ENV/', 'env.bak/', 'venv.bak/', '',
                          '# Spyder project settings', '.spyderproject', '.spyproject', '',
                          '# Rope project settings', '.ropeproject', '',
                          '# mkdocs documentation',
                          '/site', '',
                          '# mypy',
                          '.mypy_cache/', '.dmypy.json', 'dmypy.json',
                          '# Pyre type checker', '.pyre/'] + lib_gitignores)


def _check_gitignore(repo: Repo, force=False):
    """
    check if file `.gitignore`  have the needed ignored items for lumo.
    """
    # version_mark = os.path.join(repo.working_dir, FN.VERSION)
    # ignorefn = os.path.join(repo.working_dir, '.gitignore')

    # if os.path.exists(version_mark) and not force:
    #     if os.path.exists(ignorefn):
    #         return False

    old_marks = [f for f in os.listdir(repo.working_dir) if f.startswith('.lumo.')]
    for old_mark in old_marks:
        mark_fn = os.path.join(repo.working_dir, old_mark)
        if os.path.isfile(mark_fn):
            os.remove(mark_fn)

    with open(version_mark, 'w') as w:
        pass

    if not os.path.exists(ignorefn):
        with open(ignorefn, 'w', encoding='utf-8') as w:
            w.write(py_gitignore)
        return True
    else:
        amend = False
        with open(ignorefn, 'r', encoding='utf-8') as r:
            lines = [i.strip() for i in r.readlines()]
            for item in lib_gitignores:
                if item not in lines:
                    lines.append(item)
                    amend = True
        if amend:
            with open(ignorefn, 'w', encoding='utf-8') as w:
                w.write('\n'.join(lines))

    return amend


def git_config_syntax(value: str) -> str:
    """
    syntax value(especially file path) which will be stored in git config
    """
    return value.replace('\\\\', '/').replace('\\', '/')


def init_repo(dir='./') -> Tuple[Repo, bool]:
    """
    initialize a directory, including git init, lumo config and a initial commit.
    """
    path = git_dir(dir)
    init = False
    if path is not None and compare_path(path, dir):
        repo = Repo(path)
    else:
        repo = Repo.init(dir)
        init = True

    if _check_gitignore(repo=repo, force=True):
        repo.git.add('.')
        repo.index.commit('initial commit')
        init = True

    check_have_commit(repo)
    return repo, init



"""
"""
import os
from collections import OrderedDict
from pprint import pformat
from typing import Any

from lumo.utils import safe_io as io
from lumo.proc.const import CFG
from lumo.proc.path import global_config_path


def default_dict():
    return {CFG.PATH.GLOBAL_EXP: os.path.expanduser("~/.lumo/experiments"),
            CFG.PATH.DATASET: os.path.expanduser("~/.lumo/datasets"),
            CFG.PATH.PRETRAINS: os.path.expanduser("~/.lumo/pretrains"),
            CFG.PATH.CACHE: os.path.expanduser("~/.cache/lumo")}


class Globals:
    RUNTIME = CFG.FIELD.RUNTIME
    GLOBAL = CFG.FIELD.GLOBAL

    def __init__(self):
        self._configs = OrderedDict()
        for level in Config.config_levels:
            self._configs[level] = Config(level)

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, value)
        else:
            self.__setitem__(name, value)

    def __getattr__(self, item):
        return self.__getitem__(item)

    def __getitem__(self, item):
        for config in self._configs.values():
            if item in config:
                return config[item]
        return None

    def __setitem__(self, key, value):
        self._configs[CFG.FIELD.RUNTIME][key] = value

    def __repr__(self):
        return "Environ({})".format(pformat(self.items()))

    @property
    def runtime_config(self):
        return self._configs[CFG.FIELD.RUNTIME]

    @property
    def globals_config(self):
        return self._configs[CFG.FIELD.GLOBAL]

    def set(self, key, value, level=CFG.FIELD.GLOBAL):
        self._configs[level][key] = value

    def get(self, key, level=CFG.FIELD.GLOBAL, default=None):
        """"""
        return self._configs[level][key, default]

    def get_first(self, *args, default=None):
        for arg in args:
            res = self[arg]
            if res is not None:
                return res
        return default

    def items(self):
        """return all config, like dict.items()"""
        return {
            CFG.FIELD.GLOBAL: self._configs[CFG.FIELD.GLOBAL].items(),
            CFG.FIELD.RUNTIME: self._configs[CFG.FIELD.RUNTIME].items(),
        }

    def require(self, key, level=CFG.FIELD.GLOBAL):
        """
        Return value if exists, or require user to input value for the specified`key`.

        Then value be typed will be stored in config of the `level`

        Args:
            key:
            level:

        Returns:

        """
        res = self[key]
        if res is None:
            res = input(f'{level} level value `{key}` required.')
            self.add_value(key, res, level)
        return res


class Config:
    """
    Single Level Config
    """
    config_levels = [CFG.FIELD.DEFAULT, CFG.FIELD.RUNTIME, CFG.FIELD.GLOBAL]

    def __init__(self, config_level):
        assert config_level in Config.config_levels, 'config level must in {}'.format(Config.config_levels)
        self._config_level = config_level
        self._config_fn = None
        self._config_dict = None

    @property
    def config_level(self):
        return self._config_level

    @property
    def running_level(self):
        return self.config_level == CFG.FIELD.RUNTIME

    @property
    def globals_level(self):
        return self.config_level == CFG.FIELD.GLOBAL

    @property
    def default_level(self):
        return self.config_level == CFG.FIELD.DEFAULT

    @property
    def config_fn(self):
        if self.globals_level:
            self._config_fn = global_config_path()
            self._config_dict = io.load_json(self._config_fn)
        elif self.running_level:
            self._initial_os_env()
        elif self.default_level:
            self._config_dict = default_dict()
        return self._config_fn

    @property
    def config_dict(self):
        if self._config_dict is not None:
            return self._config_dict

        if self.globals_level:
            self._config_dict = io.load_json(self.config_fn)
        elif self.running_level:
            self._config_dict = {}
            self._initial_os_env()
        elif self.default_level:
            self._config_dict = default_dict()

        if self._config_dict is None:
            self._config_dict = {}

        return self._config_dict

    def _initial_os_env(self):
        for k, v in os.environ.items():
            if k.startswith('LUMO_'):
                self._config_dict[k.lstrip('LUMO_').lower()] = v

    def __setitem__(self, key, value: str):
        key = str(key).lower()
        self.config_dict[key] = value
        self._flush_config()

    def __getitem__(self, key, default=None):
        key = str(key).lower()
        return self.config_dict.get(key, default)

    def __contains__(self, item):
        return item in self.config_dict

    def __repr__(self) -> str:
        return pformat(self.items())

    def _flush_config(self):
        if self._config_fn is not None:
            io.dump_json(self.config_dict, self._config_fn)

    def get(self, key, default=None):
        return self[key, default]

    def items(self):
        return self.config_dict.items()


globs = Globals()



"""

"""
import sys

class ExitHooks(object):
    def __init__(self):
        self.exit_code = None
        self.exception = None

    def hook(self):
        self._orig_exit = sys.exit
        sys.exit = self.exit
        sys.excepthook = self.exc_handler

    def exit(self, code=0):
        with open("1","w") as f:
            f.write(str(code))
        self.exit_code = code
        self._orig_exit(code)

    def exc_handler(self, exc_type, exc, tb,*args):
        import traceback

        self.exception = exc
        print(traceback.format_exception(exc_type,exc,tb))
        # print(traceback.format_exception_only(exc_type,exc))
        # traceback.print_exception(exc_type,exc,tb)

hooks = ExitHooks()
hooks.hook()

raise KeyboardInterrupt()


# 计算 均值和标准差， copied from https://zhuanlan.zhihu.com/p/104019160
import os
import cv2
import numpy as np
from torch.utils.data import Dataset
from PIL import Image


def compute_mean_and_std(dataset):
    # 输入PyTorch的dataset，输出均值和标准差
    mean_r = 0
    mean_g = 0
    mean_b = 0

    for img, _ in dataset:
        img = np.asarray(img) # change PIL Image to numpy array
        mean_b += np.mean(img[:, :, 0])
        mean_g += np.mean(img[:, :, 1])
        mean_r += np.mean(img[:, :, 2])

    mean_b /= len(dataset)
    mean_g /= len(dataset)
    mean_r /= len(dataset)

    diff_r = 0
    diff_g = 0
    diff_b = 0

    N = 0

    for img, _ in dataset:
        img = np.asarray(img)

        diff_b += np.sum(np.power(img[:, :, 0] - mean_b, 2))
        diff_g += np.sum(np.power(img[:, :, 1] - mean_g, 2))
        diff_r += np.sum(np.power(img[:, :, 2] - mean_r, 2))

        N += np.prod(img[:, :, 0].shape)

    std_b = np.sqrt(diff_b / N)
    std_g = np.sqrt(diff_g / N)
    std_r = np.sqrt(diff_r / N)

    mean = (mean_b.item() / 255.0, mean_g.item() / 255.0, mean_r.item() / 255.0)
    std = (std_b.item() / 255.0, std_g.item() / 255.0, std_r.item() / 255.0)
    return mean, std


from typing import Union, NoReturn, overload, Dict
from torch.utils.data import DataLoader
from lumo.base_classes import TrainerStage
from lumo.kit.params import ParamsType


class DataModuleMix():
    @property
    def train_dataloader(self) -> Union[NoReturn, DataLoader]:
        raise NotImplementedError()

    @property
    def test_dataloader(self) -> Union[NoReturn, DataLoader]:
        raise NotImplementedError()

    @property
    def val_dataloader(self) -> Union[NoReturn, DataLoader]:
        raise NotImplementedError()

    @overload
    def regist_dataloader(self, train: DataLoader = None, val: DataLoader = None, test: DataLoader = None,
                          **others: Dict[str, DataLoader]):
        ...

    def regist_dataloader(self, **kwargs: dict):
        raise NotImplementedError()

    def idataloader(self, params: ParamsType, stage: TrainerStage, repeat: bool = False):
        raise NotImplementedError()


class BaseCallbackMix():
    def icallbacks(self, params: ParamsType):
        raise NotImplementedError()


class CallbackMix(BaseCallbackMix):

    def add_callback(self, callback):
        raise NotImplementedError()

    def reload_callback(self, callback):
        raise NotImplementedError()

    def remove_callback(self, callback):
        raise NotImplementedError()


class ModelMix():

    def imodels(self, params: ParamsType):
        raise NotImplementedError()

    def ioptims(self, params: ParamsType):
        raise NotImplementedError()

    def optim_state_dict(self):
        raise NotImplementedError()

    def model_state_dict(self):
        raise NotImplementedError()



"""


    提供一些提供了奇葩 feature 的类
"""
from typing import Any


class NoneItem:
    """
    Can be seen as identity element or zero element
    """

    @staticmethod
    def clone(x):
        pass

    def __eq__(self, o: object) -> bool:
        return o is None

    def __add__(self, other):
        return 0 + other

    def __mul__(self, other):
        return 1 * other

    def __sub__(self, other):
        return 0 - other

    def __truediv__(self, other):
        return 1 / other

    def __floordiv__(self, other):
        return 1 / other

    def __repr__(self):
        return "NoneItem()"

    def __lt__(self, other):
        return True

    def __le__(self, other):
        return True

    def __ge__(self, other):
        return True

    def __gt__(self, other):
        return True

    def __cmp__(self, other):
        return True

    def __ne__(self, other):
        return other is not None

    def __int__(self):
        return 0

    def __float__(self):
        return 0


class AvgItem:
    """
    用于保存累积均值的类
    avg = AvgItem()
    avg += 1 # avg.update(1)
    avg += 2
    avg += 3

    avg.item = 3 #(last item)
    avg.avg = 2 #(average item)
    avg.sum = 6
    """

    def __init__(self, weight=1) -> None:
        super().__init__()
        self._sum = 0
        self._weight = weight
        self._count = 0
        self._item = 0

    def __add__(self, other):
        self.update(other)
        return self

    def update(self, other, weight=1):
        self._sum += other * self._weight
        self._count += self._weight
        self._item = other

    @property
    def item(self):
        return self._item

    @property
    def avg(self):
        if self._count == 0:
            return 0
        return self._sum / self._count

    def __repr__(self) -> str:
        return str("{}({})".format(self._item, self.avg))

    def __getattribute__(self, name: str) -> Any:
        return super().__getattribute__(name)

    def __getattr__(self, item):
        return getattr(self._item, item)

    def __format__(self, format_spec):
        return "{{:{}}}({{:{}}})".format(format_spec, format_spec).format(self._item, self.avg)

    def __getitem__(self, item):
        return self._item[item]





"""
initialize directory, create config file, including
 - ~/.lumo/
 - ~/.cache/lumo/
"""
import os


def initialize():
    from lumo.utils.template_creator import Template
    from lumo.proc.path import libhome
    if os.path.exists(libhome()):
        return False

    (
        Template()
            .add_directory('experiments')
            .add_file('config.json', info='')
            .create(libhome())
    )

    return True



"""

"""

from lumo.utils.random import *


def test_state():
    state_dict = fix_seed(10)
    a, b, c = random.randint(0, 10), np.random.rand(1), torch.rand(1)
    set_state(state_dict)
    d, e, f = random.randint(0, 10), np.random.rand(1), torch.rand(1)

    assert a == d and b == e and c == f



"""

"""

from lumo.base_classes.attr import attr

dic = attr()


dic.b.d = "asd"
dic.a = "asd"

print(dic.items())

# print(dic["b"])
# print("g" in dic)
print(dic.hash())
# c5e72bde733e687ce89c9798f825e04c


#  code from https://github.com/DocF/Soft-NMS/blob/master/soft_nms.py
import numpy as np

def py_cpu_softnms(dets, sc, Nt=0.3, sigma=0.5, thresh=0.001, method=2):
    """
    py_cpu_softnms
    :param dets:   boexs 坐标矩阵 format [y1, x1, y2, x2]
    :param sc:     每个 boxes 对应的分数
    :param Nt:     iou 交叠门限
    :param sigma:  使用 gaussian 函数的方差
    :param thresh: 最后的分数门限
    :param method: 使用的方法
    :return:       留下的 boxes 的 index
    """

    # indexes concatenate boxes with the last column
    N = dets.shape[0]
    indexes = np.array([np.arange(N)])
    dets = np.concatenate((dets, indexes.T), axis=1)

    # the order of boxes coordinate is [y1,x1,y2,x2]
    y1 = dets[:, 0]
    x1 = dets[:, 1]
    y2 = dets[:, 2]
    x2 = dets[:, 3]
    scores = sc
    areas = (x2 - x1 + 1) * (y2 - y1 + 1)

    for i in range(N):
        # intermediate parameters for later parameters exchange
        tBD = dets[i, :].copy()
        tscore = scores[i].copy()
        tarea = areas[i].copy()
        pos = i + 1

        #
        if i != N-1:
            maxscore = np.max(scores[pos:], axis=0)
            maxpos = np.argmax(scores[pos:], axis=0)
        else:
            maxscore = scores[-1]
            maxpos = 0
        if tscore < maxscore:
            dets[i, :] = dets[maxpos + i + 1, :]
            dets[maxpos + i + 1, :] = tBD
            tBD = dets[i, :]

            scores[i] = scores[maxpos + i + 1]
            scores[maxpos + i + 1] = tscore
            tscore = scores[i]

            areas[i] = areas[maxpos + i + 1]
            areas[maxpos + i + 1] = tarea
            tarea = areas[i]

        # IoU calculate
        xx1 = np.maximum(dets[i, 1], dets[pos:, 1])
        yy1 = np.maximum(dets[i, 0], dets[pos:, 0])
        xx2 = np.minimum(dets[i, 3], dets[pos:, 3])
        yy2 = np.minimum(dets[i, 2], dets[pos:, 2])

        w = np.maximum(0.0, xx2 - xx1 + 1)
        h = np.maximum(0.0, yy2 - yy1 + 1)
        inter = w * h
        ovr = inter / (areas[i] + areas[pos:] - inter)

        # Three methods: 1.linear 2.gaussian 3.original NMS
        if method == 1:  # linear
            weight = np.ones(ovr.shape)
            weight[ovr > Nt] = weight[ovr > Nt] - ovr[ovr > Nt]
        elif method == 2:  # gaussian
            weight = np.exp(-(ovr * ovr) / sigma)
        else:  # original NMS
            weight = np.ones(ovr.shape)
            weight[ovr > Nt] = 0

        scores[pos:] = weight * scores[pos:]

    # select the boxes and keep the corresponding indexes
    inds = dets[:, 4][scores > thresh]
    keep = inds.astype(int)

    return keep



import hashlib
import random
from typing import Union

import numpy as np
import torch


def hashseed(hashitem: Union[int, str]):
    assert isinstance(hashitem, (int, str))

    if isinstance(hashitem, str):
        digest = hashlib.md5(hashitem.encode(encoding='utf-8')).digest()
        return sum([int(i) for i in digest])

    return hashitem


def fix_seed(seed=10):
    """

    Args:
        seed:

    Returns:

    Notes:
        When use dataloader and its num_workers is bigger than one, the final results may can't be the same cased by multithread.
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.random.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
    fix_cuda()
    return get_state()


def fix_cuda():
    if torch.cuda.is_available():
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.enabled = True


def get_state():
    return {
        "numpy": np.random.get_state(),
        "torch": torch.random.get_rng_state(),
        "torch.cuda": torch.cuda.get_rng_state() if torch.cuda.is_available() else None,
        "random": random.getstate(),
    }


def set_state(state_dict):
    random.setstate(state_dict["random"])
    np.random.set_state(state_dict["numpy"])
    torch.random.set_rng_state(state_dict["torch"])
    if torch.cuda.is_available():
        if "torch.cuda" in state_dict:
            torch.cuda.set_rng_state(state_dict["torch.cuda"])
        else:
            import warnings
            warnings.warn("Don't have torch.cuda random state")
    fix_cuda()



from torchvision.datasets.cifar import CIFAR100
import torch
import numpy as np
import random
import os
from lumo.proc.path import cache_dir
from lumo.utils import cacher
from lumo.kit.logger import get_global_logger

log = get_global_logger()


class MockImage:
    def __init__(self, image_size: int = 32):
        mock_dataset_root = os.path.join(cache_dir(), 'mock')
        cache_args = (MockImage, mock_dataset_root)
        item, path = cacher.load_if_exists(*cache_args)
        self.image_size = image_size
        if item is not None:
            log.dddebug(f'Load MockImage from cache {path}')
            self.data, self.targets = item
        else:
            cifar100 = CIFAR100(root=mock_dataset_root, train=True, download=True)
            self.data, self.targets = cifar100.data, cifar100.targets
            path = cacher.save_cache([cifar100.data, cifar100.targets], *cache_args)
            log.dddebug(f'Save MockImage cache to {path}')

    def image(self, channle_first=False):
        res = random.choice(self.data)  # type: np.ndarray
        if channle_first:
            res = res.transpose(2, 0, 1)
        res = torch.from_numpy(res)

        return res

    def images(self, bs=2, channle_first=False):
        ind = np.random.permutation(len(self.data))[:bs]
        res = self.data[ind]
        if channle_first:
            res = res.transpose(0, 3, 1, 2)
        res = torch.from_numpy(res)

        return res



"""

"""
import sys
sys.path.insert(0,"../")
from lumo import __version__
print(__version__)


from lumo import Meter,AvgMeter
import torch

m = Meter()
m.a = 1
m.b = "2"
m.c = torch.rand(1)[0]

m.c1 = torch.rand(1)
m.c2 = torch.rand(2)
m.c3 = torch.rand(4, 4)
print(m)

m = Meter()
m.a = 0.236
m.b = 3.236
m.c = 0.23612312
m.percent(m.a_)
m.int(m.b_)
m.float(m.c_,2)
print(m)


am = AvgMeter()
for j in range(5):
    for i in range(100):
        m = Meter()
        m.percent(m.c_)
        m.a = 1
        m.b = "2"
        m.c = torch.rand(1)[0]

        m.c1 = torch.rand(1)
        m.c2 = torch.rand(2)
        m.c3 = torch.rand(4, 4)
        m.d = [4]
        m.e = {5: "6"}
        # print(m)
        am.update(m)
    print(am)

from lumo import Meter
m = Meter()
print(m.k)

m.all_loss = m.all_loss + 5
m.all_loss = m.all_loss + 3
print(m.all_loss)


from typing import Optional, Sequence

from torch.utils.data import dataloader as _loader, Dataset, Sampler
from torch.utils.data.dataloader import T_co, _collate_fn_t, _worker_init_fn_t


class DataLoader(_loader.DataLoader):
    def __init__(self, dataset: Dataset[T_co], batch_size: Optional[int] = 1, shuffle: bool = False,
                 sampler: Optional[Sampler[int]] = None, batch_sampler: Optional[Sampler[Sequence[int]]] = None,
                 num_workers: int = 0, collate_fn: _collate_fn_t = None, pin_memory: bool = False,
                 drop_last: bool = False, timeout: float = 0, worker_init_fn: _worker_init_fn_t = None,
                 multiprocessing_context=None, generator=None, *, prefetch_factor: int = 2,
                 persistent_workers: bool = False):
        super().__init__(dataset, batch_size, shuffle, sampler, batch_sampler, num_workers, collate_fn, pin_memory,
                         drop_last, timeout, worker_init_fn, multiprocessing_context, generator,
                         prefetch_factor=prefetch_factor, persistent_workers=persistent_workers)

    def _get_iterator(self) -> _loader._BaseDataLoaderIter:
        return super()._get_iterator()

    @property
    def multiprocessing_context(self):
        return super().multiprocessing_context()

    def __setattr__(self, attr, val):
        return super().__setattr__(attr, val)

    def __iter__(self) -> _loader._BaseDataLoaderIter:
        return super().__iter__()

    @property
    def _auto_collation(self):
        return super()._auto_collation()

    @property
    def _index_sampler(self):
        return super()._index_sampler()

    def __len__(self) -> int:
        return super().__len__()



from enum import Enum


class TrainerStage(Enum):
    debug = -2
    init = -1
    train = 0
    train_epoch = 1
    test = 2
    val = 3

    @property
    def is_train(self):
        return self == self.train

    @property
    def is_test(self):
        return self == self.test

    @property
    def is_eval(self):
        return self == self.val



"""

"""

from lumo.utils.timing import TimeIt
import time

if __name__ == '__main__':
    t = TimeIt()
    t.start()
    for i in range(10):
        time.sleep(0.1)
        t.mark(i)

    t.end()
    print(t.meter())
from torch.utils.data.dataloader import DataLoader

from torch.nn import Module


"""

"""

from lumo.base_classes.trickitems import NoneItem

none = NoneItem()
assert none > 0
assert none >= 0
assert none < 0
assert none <= 0
assert none != 0
assert none == None
assert none is not None
assert none/2 == 0.5
assert none*2 == 2
assert none+1 == 1
assert none-1 == -1




"""
Methods about screen single line outputs.
"""
import sys
import time
import collections
from lumo.proc.console import get_consolo_width, support_multiline
import numpy as np


class ScreenStr:
    """
    A ScreenStr start with '\r' won't overflow, any string outside the screen width will be cut.

    Notes:
    If output consolo support multiline(like pycharm or jupyter notebook) return, all string will be represented.
    """
    t = 0
    dt = 0.7
    last = 0
    left = 0
    max_wait = 1.
    wait = 0
    wait_toggle = False

    debug = False
    last_width = 0
    multi_mode = support_multiline()

    def __init__(self, content="", leftoffset=0) -> None:
        self.content = content
        ScreenStr.left = leftoffset

    def __repr__(self) -> str:
        if ScreenStr.multi_mode:
            return self.content
        return self._screen_str()

    def tostr(self):
        return self.content

    @classmethod
    def set_speed(cls, dt: float = 0.05):
        cls.dt = dt

    @classmethod
    def deltatime(cls):
        if cls.last == 0:
            cls.last = time.time()
            return 0
        else:
            end = time.time()
            res = end - cls.last
            cls.last = end
            return res

    @classmethod
    def cacu_offset_(cls, out_width):

        delta = cls.deltatime()
        cls.t += delta * cls.dt

        # pi = 2*math.pi
        t = cls.t
        # k = 2 * out_width / pi
        k = 10
        pi = 2 * out_width / k
        offset = round(k * (t % pi) * ((t % pi) < pi / 2) + (-k * (t % pi) + 2 * out_width) * ((t % pi) > pi / 2))

        # offset = math.floor(out_width * (math.cos(ScreenStr.t + math.pi) + 1) / 2)
        # print(offset)
        return offset

    a = 1

    def __len__(self) -> int:
        txt = self.content.encode("gbk", errors='ignore')
        return len(txt)

    def _decode_sub(self, txt, left, right):
        try:
            txt = txt[left:right].decode("gbk", errors='ignore')
        except:
            try:
                txt = txt[left:right - 1].decode("gbk", errors='ignore')
            except:
                try:
                    txt = txt[left + 1:right].decode("gbk", errors='ignore')
                except:
                    txt = txt[left + 1:right - 1].decode("gbk", errors='ignore')

        return txt

    @staticmethod
    def consolo_width():
        width = get_consolo_width()
        return width

    @staticmethod
    def split(txt, len):
        try:
            return txt[:len], txt[len:]
        except:
            try:
                return txt[:len + 1], txt[len + 1:]
            except:
                return txt[:len - 1], txt[len - 1:]

    def _screen_str(self, margin="..."):
        width = self.consolo_width()

        txt = self.content.encode("gbk", errors='ignore').strip()
        textlen = len(txt)

        if textlen <= width:
            return self.content

        left, right = self.split(txt, self.left)
        if len(left) >= width:
            return left[:width]

        offset = 0  # TODO

        offright = width - len(left) + offset - len(margin)

        left = left.decode("gbk", errors='ignore')
        right = self._decode_sub(right, offset, offright)

        head = "\r" if self.content.startswith("\r") else ""
        tail = "\n" if self.content.endswith("\n") else ""

        txt = "{}{}{}{}".format(head, left, right, tail)
        return txt + margin


class Progbar(object):
    """从 keras 中扒下来的进度条显示类

    Displays a progress bar.
    Arguments
        target: Total number of steps expected, None if unknown.
        width: Progress bar width on screen.
        verbose: Verbosity mode, 0 (silent), 1 (verbose), 2 (semi-verbose)
        stateful_metrics: Iterable of string names of metrics that
            should *not* be averaged over time. Metrics in this list
            will be displayed as-is. All others will be averaged
            by the progbar before display.
        interval: Minimum visual progress update interval (in seconds).
    """

    def __init__(self, target, width=30, verbose=1, interval=0.05,
                 stateful_metrics=None):
        self.target = target
        self.width = width
        self.verbose = verbose
        self.interval = interval
        if stateful_metrics:
            self.stateful_metrics = set(stateful_metrics)
        else:
            self.stateful_metrics = set()

        self._dynamic_display = ((hasattr(sys.stdout, 'isatty') and
                                  sys.stdout.isatty()) or
                                 'ipykernel' in sys.modules)
        self._total_width = 0
        self._seen_so_far = 0
        self._values = collections.OrderedDict()
        self._start = time.time()
        self._last_update = 0

    def update(self, current, values=None):
        """
        Updates the progress bar.

        Arguments
            current: Index of current step.
            values: List of tuples:
                `(name, value_for_last_step)`.
                If `name` is in `stateful_metrics`,
                `value_for_last_step` will be displayed as-is.
                Else, an average of the metric over time will be displayed.
        """
        strs = []
        values = values or []
        for k, v in values:
            if k not in self.stateful_metrics:
                if k not in self._values:
                    self._values[k] = [v * (current - self._seen_so_far),
                                       current - self._seen_so_far]
                else:
                    self._values[k][0] += v * (current - self._seen_so_far)
                    self._values[k][1] += (current - self._seen_so_far)
            else:
                # Stateful metrics output a numeric value.  This representation
                # means "take an average from a single value" but keeps the
                # numeric formatting.
                self._values[k] = [v, 1]
        self._seen_so_far = current

        now = time.time()
        info = ' - %.0fs' % (now - self._start)
        if self.verbose == 1:
            if (now - self._last_update < self.interval and
                    self.target is not None and current < self.target):
                return

            prev_total_width = self._total_width

            if self.target is not None:
                numdigits = int(np.floor(np.log10(self.target))) + 1
                barstr = '%%%dd/%d [' % (numdigits, self.target)
                bar = barstr % current
                prog = float(current) / self.target
                prog_width = int(self.width * prog)
                if prog_width > 0:
                    bar += ('=' * (prog_width - 1))
                    if current < self.target:
                        bar += '>'
                    else:
                        bar += '='
                bar += ('.' * (self.width - prog_width))
                bar += ']'
            else:
                bar = '%7d/Unknown' % current

            self._total_width = len(bar)
            strs.append(bar)

            if current:
                time_per_unit = (now - self._start) / current
            else:
                time_per_unit = 0
            if self.target is not None and current < self.target:
                eta = time_per_unit * (self.target - current)
                if eta > 3600:
                    eta_format = ('%d:%02d:%02d' %
                                  (eta // 3600, (eta % 3600) // 60, eta % 60))
                elif eta > 60:
                    eta_format = '%d:%02d' % (eta // 60, eta % 60)
                else:
                    eta_format = '%ds' % eta

                info = ' - ETA: %s' % eta_format
            else:
                if time_per_unit >= 1:
                    info += ' %.0fs/step' % time_per_unit
                elif time_per_unit >= 1e-3:
                    info += ' %.0fms/step' % (time_per_unit * 1e3)
                else:
                    info += ' %.0fus/step' % (time_per_unit * 1e6)

            for k in self._values:
                info += ' - %s:' % k
                if isinstance(self._values[k], list):
                    avg = np.mean(
                        self._values[k][0] / max(1, self._values[k][1]))
                    if abs(avg) > 1e-3:
                        info += ' %.4f' % avg
                    else:
                        info += ' %.4e' % avg
                else:
                    info += ' %s' % self._values[k]

            self._total_width += len(info)
            if prev_total_width > self._total_width:
                info += (' ' * (prev_total_width - self._total_width))

            if self.target is not None and current >= self.target:
                info += '\n'

            strs.append(info)

        elif self.verbose == 2:
            if self.target is None or current >= self.target:
                for k in self._values:
                    info += ' - %s:' % k
                    avg = np.mean(
                        self._values[k][0] / max(1, self._values[k][1]))
                    if avg > 1e-3:
                        info += ' %.4f' % avg
                    else:
                        info += ' %.4e' % avg
                info += '\n'

                strs.append(info)

        self._last_update = now

        return "".join(strs)

    def add(self, n, values=None):
        self.update(self._seen_so_far + n, values)



"""

"""
from collections import OrderedDict
from itertools import cycle, chain
from typing import Optional, overload, Union
import torch
from lumo.contrib.itertools import safe_cycle


class DataBundler:
    """
    """

    class DataBundler(OrderedDict):
        pass

    def __init__(self, device=None):
        self.dataloaders = DataBundler.DataBundler()
        self.iter_mode = "chain"
        self.device_arg = None
        if device is not None:
            self.to(device)

    def __len__(self):
        if self.iter_mode == 'zip':
            return min(
                len(loader)
                for name, (loader, func) in self.dataloaders.items()
                if func.__name__ != 'safe_cycle'
            )
        elif self.iter_mode == "chain":
            return sum(self.len_list())

    def __getitem__(self, item):
        return self.dataloaders[item][0]

    def __iter__(self):
        loaders = self._func_loader()
        if len(loaders) == 1:
            iter = loaders[0]

        elif self.iter_mode == "zip":
            iter = zip(*loaders)
        elif self.iter_mode == "chain":
            iter = chain(*loaders)
        else:
            assert False

        for batch_data in iter:
            yield batch_data

    def _append(self, loader, func, name):
        from torch.utils.data import DataLoader
        assert isinstance(loader, (DataLoader, DataBundler))
        if name is None:
            unname = "unnamed"
            i = 0
            name = "{}_{}".format(unname, i)
            while name in self.dataloaders:
                i += 1
                name = "{}_{}".format(unname, i)
        else:
            assert name not in self.dataloaders, "{} also defined in bundler".format(name)

        self.dataloaders[name] = (loader, func)
        if isinstance(loader, DataBundler) and self.device_arg is not None:
            loader.to(*self.device_arg[0], **self.device_arg[1])

    def _func_loader(self):
        def filter(loader):
            if isinstance(loader, DataBundler):
                return iter(loader)
            return loader

        return [func(filter(loader)) for name, (loader, func) in self.dataloaders.items()]

    def set_batch_size(self, batch_size):
        from torch.utils.data import DataLoader
        from thexp.contrib.data.dataloader import DataLoader as thDataLoader
        for _, (loader, _) in self.dataloaders.items():
            if isinstance(loader, thDataLoader):
                loader.set_batch_size(batch_size)
            elif isinstance(loader, DataLoader):
                loader.batch_sampler.batch_size = batch_size

    def len_list(self):
        """
        按照添加的顺序返回各个dataloader的长度（batch级别）
        :return:
        """
        return [len(loader) for _, (loader, _) in self.dataloaders.items()]

    def len_dict(self):
        """
        返回每个loader的 name:len 字典
        :return: an OrderedDict
        """
        res = DataBundler.DataBundler()
        for name, (loader, func) in self.dataloaders.items():
            res[name] = len(loader)
        return res

    def cycle(self, loader, name=None):
        """

        :param loader: Dataloader object
        :param name:
        :return:
        """
        """一般在zip中保证数据量少的数据集不会成为拖累"""
        self._append(loader, safe_cycle, name)
        return self

    def add(self, loader, name=None):
        self._append(loader, lambda x: x, name)
        return self

    @overload
    def to(self, dtype: torch.dtype, non_blocking: bool = False, copy: bool = False):
        ...

    @overload
    def to(self, device: Optional[Union[torch.device, str]] = None, dtype: Optional[torch.dtype] = None,
           non_blocking: bool = False, copy: bool = False):
        ...

    @overload
    def to(self, other: torch.Tensor, non_blocking: bool = False, copy: bool = False):
        ...

    def to(self, *args, **kwargs):
        if len(args) == 0:
            assert 'device' in kwargs
            device = kwargs['device']
        else:
            device = args[0]
        assert isinstance(device, (torch.device, str))
        self.device_arg = [args, kwargs]
        for (loader, _) in self.dataloaders.values():
            if isinstance(loader, DataBundler):
                loader.to(*args, **kwargs)
        return self

    def choice_batch(self) -> tuple:
        return next(iter(self))

    def zip_mode(self):
        """切换为zip方式提供所有添加的数据集"""
        self.iter_mode = "zip"
        return self

    def chain_mode(self):
        """
        切换为chain方式提供所有添加的数据集
            注意，如果以cycle方法添加了某个数据集，那么该迭代将永远不会停止
        :return:
        """
        self.iter_mode = "chain"
        return self

    @staticmethod
    def create_zip_bundler(**kwargs):
        bundler = DataBundler()
        loaders = [(len(v), v, k) for k, v in kwargs.items()]
        loaders.sort(reverse=True)
        bundler.add(loaders[0][1], loaders[0][2])
        for (_, loader, name) in loaders[1:]:
            bundler.cycle(loader, name)
        return bundler

    @staticmethod
    def create_chain_bundler(*args):
        bundler = DataBundler()
        for loader in args:
            bundler.add(loader)
        return bundler

    def __repr__(self):
        from pprint import pformat
        return pformat(self.len_dict())



"""

"""
from typing import Tuple, List, Callable

import torch


def rotate_right_angle(x: torch.Tensor, w_dim: int = 2, h_dim: int = 3, degree: int = 90):
    assert degree in {90, 270, 180}
    if degree == 90:
        x = x.transpose(w_dim, h_dim)  # 90
    elif degree == 180:
        x = x.flip(w_dim)
    elif degree == 270:
        x = x.transpose(w_dim, h_dim).flip(h_dim)  # 270

    return x


def split_sub_matrix(mat: torch.Tensor, *sizes):
    """
    将一个[N,M,...,L]的矩阵按 n,m,...l 拆分成 N/n*M/m*...L/l 个 [n,m,...l]的小矩阵

    如果N/n 无法整除，不会报错而是会将多余的裁掉
    example:
        mat = torch.arange(0,24).view(4,6) # shape = [4, 6]
        >> tensor([[ 0,  1,  2,  3,  4,  5],
                   [ 6,  7,  8,  9, 10, 11],
                   [12, 13, 14, 15, 16, 17],
                   [18, 19, 20, 21, 22, 23]])

        split_sub_matrix(mat,2,3) # shape = [2, 2, 2, 3]
        >> tensor([[[[ 0,  1,  2],
                      [ 6,  7,  8]],

                     [[ 3,  4,  5],
                      [ 9, 10, 11]]],

                    [[[12, 13, 14],
                      [18, 19, 20]],

                     [[15, 16, 17],
                      [21, 22, 23]]]])

    :param mat: 一个[N,M,...L] 的矩阵
    :param sizes: n,m,...l 的list, 其长度不一定完全和mat的维数相同
        mat = torch.arange(0,240).view([4,6,10])
        split_sub_matrix(mat,2,3) # shape = [2, 2, 10, 2, 3]
    :return: 一个 [N/row,M/col,row,col] 的矩阵
    """
    for i, size in enumerate(sizes):
        mat = mat.unfold(i, size, size)
    return mat


def onehot(labels: torch.Tensor, label_num):
    """
    convert label to onehot vector
    Args:
        labels:
        label_num:

    Returns:

    """
    return torch.zeros(labels.shape[0], label_num, device=labels.device).scatter_(1, labels.view(-1, 1), 1)


def cartesian_product(left: torch.Tensor, right: torch.Tensor = None) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Calculate cartesian product of the given tensor(s)

    Args:
        left: A pytorch tensor.
        right: A pytorch tensor,
            if None, wile be left X left

    Returns:
        Tuple[torch.Tensor, torch.Tensor]

    Example:
        >>> cartesian_product(torch.arange(0,3),torch.arange(0,5))

    (tensor([0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2]),
      tensor([0, 1, 2, 3, 4, 0, 1, 2, 3, 4, 0, 1, 2, 3, 4]))

    """
    if right is None:
        right = left

    nleft = left.repeat_interleave(right.shape[0], dim=0)
    nright = right.repeat(*[item if i == 0 else 1 for i, item in enumerate(left.shape)])
    return nleft, nright


def cat_then_split(op: Callable[[torch.Tensor], torch.Tensor], tensors: List[torch.Tensor]) -> List[torch.Tensor]:
    """
    Examples:
    >>> ta,tb,tc = cat_then_split(lambda x:model(x),[ta,tb,tc])
    >>> alogits,blogits,clogits = cat_then_split(model,[ta,tb,tc])
    """

    res = op(torch.cat(tensors))  # type: torch.Tensor
    return res.split_with_sizes([i.shape[0] for i in tensors])


def label_smoothing(onehot_labels, epsilon=0.1):
    """
    Applies label smoothing


    Args:
        onehot_labels:
        epsilon:

    Returns:

    References:
        see
        https://arxiv.org/abs/1512.00567
        https://arxiv.org/abs/1906.02629
    """
    return ((1 - epsilon) * onehot_labels) + (epsilon / onehot_labels.shape[-1])


def label_smoothing_v2(onehot_labels, rate=0.1):
    """
    Applies label smoothing


    Args:
        onehot_labels:
        epsilon:

    Returns:

    References:
        see
        https://arxiv.org/abs/1512.00567
        https://arxiv.org/abs/1906.02629
    """
    mixed = torch.ones_like(onehot_labels) * (1 / onehot_labels.shape[-1])
    return ((1 - rate) * onehot_labels) + rate * mixed


def elementwise_mul(vec: torch.Tensor, mat: torch.Tensor) -> torch.Tensor:
    """
    Element-wise multiplication of a vector and a matrix

    References:
        https://discuss.pytorch.org/t/element-wise-multiplication-of-a-vector-and-a-matrix/56946

    Args:
        vec: tensor of shape (N, )
        mat: matrix of shape (N, ...)

    Returns:
        A tensor

    """
    nshape = [-1] + [1] * (len(mat.shape) - 1)
    nvec = vec.reshape(nshape)  # .expand_as(mat)
    return nvec * mat


def euclidean_dist(x, y, min=1e-12):
    """
    copy from https://blog.csdn.net/IT_forlearn/article/details/100022244
    Args:
      x: pytorch Variable, with shape [m, d]
      y: pytorch Variable, with shape [n, d]
    Returns:
      dist: pytorch Variable, with shape [m, n]
    """

    m, n = x.size(0), y.size(0)
    # xx经过pow()方法对每单个数据进行二次方操作后，在axis=1 方向（横向，就是第一列向最后一列的方向）加和，此时xx的shape为(m, 1)，经过expand()方法，扩展n-1次，此时xx的shape为(m, n)
    xx = torch.pow(x, 2).sum(1, keepdim=True).expand(m, n)
    # yy会在最后进行转置的操作
    yy = torch.pow(y, 2).sum(1, keepdim=True).expand(n, m).t()
    dist = xx + yy
    # torch.addmm(beta=1, input, alpha=1, mat1, mat2, out=None)，这行表示的意思是dist - 2 * x * yT
    dist = dist - 2 * torch.mm(x, y.T)
    # dist.addmm_(1, -2, x, y.t())
    # clamp()函数可以限定dist内元素的最大最小范围，dist最后开方，得到样本之间的距离矩阵
    dist = dist.clamp(min=min).sqrt()  # for numerical stability
    return dist


def label_eq_matric(labels: torch.Tensor):
    """

    Args:
        labels: ground truth of shape [bsz].

    Returns:
        a bool tensor [bsz, bsz],
    """
    labels = labels.contiguous().view(-1, 1)
    mask = torch.eq(labels, labels.T)

    return mask


def resize(arr, newsize, mode='bilinear'):
    pass  # TODO



# /usr/bin/python3

import os

dist_dir = os.path.join(os.path.dirname(__file__),'dist')
fs = os.listdir(dist_dir)

fs = [os.path.join(i) for i in fs]
fs = sorted(fs, key=lambda x: os.path.getatime(os.path.join(dist_dir,x)), reverse=False)

print(fs[-1])




"""

"""

import sys
sys.path.insert(0,"../../")
import time

from lumo.utils.screen import ScreenStr
s = "\rLong Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text;Long Text"
print(ScreenStr(s,leftoffset=10),end="")
for i in range(100):
    time.sleep(0.2)


from lumo.contrib.data.collate import AutoCollate
from torch.utils.data.dataloader import DataLoader
import torch
device = torch.device('cuda:1')
DataLoader(...,collate_fn=AutoCollate(device))



import os
from typing import Optional, List, Union
from packaging import version

import torch
from accelerate import Accelerator as HugAccelerator

from accelerate.kwargs_handlers import DistributedDataParallelKwargs, GradScalerKwargs, KwargsHandler
from accelerate.utils import RNGType, extract_model_from_parallel, gather, pad_across_processes, save, wait_for_everyone

from accelerate.state import AcceleratorState as HugAcceleratorState


class AcceleratorState(HugAcceleratorState):

    def __init__(self, fp16: bool = None, cpu: bool = False, _from_accelerator: bool = False, device=None):
        super().__init__(fp16, cpu, _from_accelerator)
        self.cpu = cpu
        self.to_device(device)

    def to_device(self, device):
        """device can be changed only in single GPU mode"""
        if device is not None:
            if int(os.environ.get("LOCAL_RANK", -1)) == -1 and not self.cpu:
                self.device = torch.device(device)
                return True
        return False


class Accelerator(HugAccelerator):
    def __init__(self, device_placement: bool = True, split_batches: bool = False, fp16: bool = None, cpu: bool = False,
                 rng_types: Optional[List[Union[str, RNGType]]] = None,
                 kwargs_handlers: Optional[List[KwargsHandler]] = None, device=None,
                 accelerator_state: HugAcceleratorState = None):
        super().__init__(device_placement, split_batches, fp16, cpu, rng_types, kwargs_handlers)
        if accelerator_state is None:
            self.state = AcceleratorState(fp16=fp16, cpu=cpu, _from_accelerator=True, device=device)
        else:
            self.state = accelerator_state

        self.device_placement = device_placement
        self.split_batches = split_batches

        # Kwargs handlers
        self.ddp_handler = None
        self.scaler_handler = None
        if kwargs_handlers is not None:
            for handler in kwargs_handlers:
                assert isinstance(handler, KwargsHandler), f"Unsupported kwargs handler passed: {handler}."
                if isinstance(handler, DistributedDataParallelKwargs):
                    if self.ddp_handler is not None:
                        raise ValueError("You can only pass one `DistributedDataParallelKwargs` in `kwargs_handler`.")
                    else:
                        self.ddp_handler = handler
                elif isinstance(handler, GradScalerKwargs):
                    if self.scaler_handler is not None:
                        raise ValueError("You can only pass one `GradScalerKwargs` in `kwargs_handler`.")
                    else:
                        self.scaler_handler = handler

        # Mixed precision attributes
        self.scaler = None
        self.native_amp = False
        if self.state.use_fp16:
            self.native_amp = version.parse(torch.__version__) >= version.parse("1.6")
            kwargs = self.scaler_handler.to_kwargs() if self.scaler_handler is not None else {}
            self.scaler = torch.cuda.amp.GradScaler(**kwargs)

        # Internal references to the training objects
        self._optimizers = []

        # RNG Types
        if rng_types is None:
            self.rng_types = ["torch"] if version.parse(torch.__version__) <= version.parse("1.5.1") else ["generator"]

    def prepare_data_loader(self, data_loader):
        return super().prepare_data_loader(data_loader)




#! /usr/bin/python

# Voice Activity Detection (VAD) tool.
# use the vad_help() function for instructions.
# Navid Shokouhi December 2012.

# Updated: May 2017 for Speaker Recognition collaboration.

from .audio_tools import *
import numpy as np


#### Display tools
def plot_this(s, title=''):
    """
     
    """
    import pylab
    s = s.squeeze()
    if s.ndim == 1:
        pylab.plot(s)
    else:
        pylab.imshow(s, aspect='auto')
        pylab.title(title)
    pylab.show()


def plot_these(s1, s2):
    import pylab
    try:
        # If values are numpy arrays
        pylab.plot(s1 / max(abs(s1)), color='red')
        pylab.plot(s2 / max(abs(s2)), color='blue')
    except:
        # Values are lists
        pylab.plot(s1, color='red')
        pylab.plot(s2, color='blue')
    pylab.legend()
    pylab.show()


#### Energy tools
def zero_mean(xframes):
    """
        remove mean of framed signal
        return zero-mean frames.
        """
    m = np.mean(xframes, axis=1)
    xframes = xframes - np.tile(m, (xframes.shape[1], 1)).T
    return xframes


def compute_nrg(xframes):
    # calculate per frame energy
    n_frames = xframes.shape[1]
    return np.diagonal(np.dot(xframes, xframes.T)) / float(n_frames)


def compute_log_nrg(xframes):
    # calculate per frame energy in log
    n_frames = xframes.shape[1]
    raw_nrgs = np.log(compute_nrg(xframes + 1e-5)) / float(n_frames)
    return (raw_nrgs - np.mean(raw_nrgs)) / (np.sqrt(np.var(raw_nrgs)))


def power_spectrum(xframes):
    """
        x: input signal, each row is one frame
        """
    X = np.fft.fft(xframes, axis=1)
    X = np.abs(X[:, :X.shape[1] / 2]) ** 2
    return np.sqrt(X)


def nrg_vad(xframes, percent_thr, nrg_thr=0., context=5):
    """
        Picks frames with high energy as determined by a 
        user defined threshold.
        
        This function also uses a 'context' parameter to
        resolve the fluctuative nature of thresholding. 
        context is an integer value determining the number
        of neighboring frames that should be used to decide
        if a frame is voiced.
        
        The log-energy values are subject to mean and var
        normalization to simplify the picking the right threshold. 
        In this framework, the default threshold is 0.0
        """
    xframes = zero_mean(xframes)
    n_frames = xframes.shape[0]

    # Compute per frame energies:
    xnrgs = compute_log_nrg(xframes)
    xvad = np.zeros((n_frames, 1))
    for i in range(n_frames):
        start = max(i - context, 0)
        end = min(i + context, n_frames - 1)
        n_above_thr = np.sum(xnrgs[start:end] > nrg_thr)
        n_total = end - start + 1
        xvad[i] = 1. * ((float(n_above_thr) / n_total) > percent_thr)
    return xvad


def vad_x(file):
    fs, ns, s = read_wav(test_file)
    win_len = int(fs * 0.025)
    hop_len = int(fs * 0.010)
    sframes = enframe(ns, win_len, hop_len)  # rows: frame index, cols: each frame
    percent_high_nrg = 0.5
    vad = nrg_vad(sframes, percent_high_nrg)
    x_samples = deframe(vad, win_len, hop_len)
    return ns[np.where(x_samples.squeeze().astype(np.int) == 1)]



if __name__ == '__main__':
    # test_file='data/sa1.wav'
    test_file = 'data/L001-train-000.wav'
    fs, ns, s = read_wav(test_file)
    win_len = int(fs * 0.025)
    hop_len = int(fs * 0.010)
    sframes = enframe(ns, win_len, hop_len)  # rows: frame index, cols: each frame
    plot_this(compute_log_nrg(sframes))

    # percent_high_nrg is the VAD context ratio. It helps smooth the
    # output VAD decisions. Higher values are more strict.
    percent_high_nrg = 0.5

    vad = nrg_vad(sframes, percent_high_nrg)
    x_samples = deframe(vad, win_len, hop_len)
    plot_these(x_samples, ns)
    from matplotlib import pyplot as plt

    plt.plot(ns[np.where(x_samples.squeeze().astype(np.int) == 1)])
    plt.show()



"""

"""
import sys
sys.path.insert(0,"../")
from lumo import __version__
print(__version__)
import time

from lumo import Experiment

exp = Experiment("expname")
print(exp.make_exp_dir("explevel"))
print(exp.makedir("testlevel"))

from lumo import Params

params = Params()
exp.add_plugin("params", dict(
    _param_hash = params.hash(),
    data = params.inner_dict().jsonify()
))

from lumo import Logger

logger = Logger()
fn = logger.add_log_dir(exp.makedir('logger'))
exp.add_plugin('logger', dict(
    fn = fn,
))



time.sleep(1)
try:
    raise Exception("dddd")
except:
    pass

exp.end()





from lumo.base_classes import attr
import numpy as np
import torch


class NAttr(attr):

    def __init__(self) -> None:
        super().__init__()
        self.k = None


def get_res():
    res = attr()
    res.a = 1
    res.nn = None
    res.kk = NAttr()
    res.st = 'NAttr()'
    res.b = [2, 3, 4]
    res.c = {'a': 1, 'b': [5, 6, 7], 'c': {'d': [8, 9]}}
    res.d = torch.tensor(1).float()
    res.e = torch.tensor([2, 3, 4]).float()
    res.f = np.array(2)
    res.g = np.array([2, 3, 4])
    return res


def test_jsonify():
    res = get_res()
    jsn = res.jsonify()
    rres = attr().from_dict(jsn)

    assert rres.hash() == res.hash()


def test_copy():
    res = get_res()
    copy = res.copy()

    res.a = 2
    res.b.append(5)
    res.c.a = 2
    res.c.b.append(8)
    res.c.c.d.append(8)

    res.d.uniform_()
    res.e.uniform_()

    assert copy.a == 1
    assert copy.b == [2, 3, 4]
    assert copy.c.a == 1
    assert copy.c.b == [5, 6, 7]
    assert copy.c.c.d == [8, 9]
    assert copy.d == torch.tensor(1)
    assert (copy.e == torch.tensor([2, 3, 4])).all()


def test_deepcopy():
    res = get_res()
    from copy import deepcopy
    copy = deepcopy(res)
    assert res.hash() == copy.hash()


def test_replace():
    res = get_res()
    res.replace(a=6).replace(b=7)
    assert res.a == 6
    assert res.b == 7



import os
import sys


def cache_dir():
    res = os.path.expanduser("~/.cache/lumo")
    os.makedirs(res, exist_ok=True)
    return res


def dataset_cache_dir(name=None):
    if name is None:
        return os.path.join(cache_dir(), 'datasets')
    else:
        return os.path.join(cache_dir(), 'datasets', name)


def libhome():
    return os.path.expanduser("~/.lumo")


def git_dir(root='./'):
    """
    git repository directory
    Args:
        root:

    Returns:

    """
    from lumo.proc.explore import git_enable
    if git_enable():
        from git import Git
        cur = os.getcwd()
        os.chdir(root)
        res = Git().execute(['git', 'rev-parse', '--git-dir'])
        res = os.path.abspath(os.path.dirname(res))
        os.chdir(cur)
        return res
    else:
        return None


def local_dir():
    res = git_dir()
    if res is None:
        res = os.path.dirname(os.path.join(os.getcwd(), sys.argv[0]))
    return res


def global_config_path():
    return os.path.join(libhome(), "config.json")



import os


def split_file(file, line_chunk=1000, fmt='{:02d}'):
    """
    split one text file
        <filename>.<suf>
    into
        <filename>_<fmt_idx>.<suf>
    Returns:

    """
    c = 0

    pass



"""

"""
from typing import Union, List, Dict, Tuple

import torch


def to_device(batch: Union[List, Dict, Tuple, torch.Tensor], device_args_kargs):
    if isinstance(batch, (list, tuple)):
        return [to_device(ele, device_args_kargs) for ele in batch]
    elif isinstance(batch, dict):
        return {k: to_device(ele, device_args_kargs) for k, ele in batch.items()}
    elif isinstance(batch, torch.Tensor):
        return batch.to(*device_args_kargs[0], **device_args_kargs[1])



from datetime import datetime
import os
import shutil
from collections import OrderedDict
from dataclasses import dataclass
from functools import lru_cache
from typing import Dict, List

from git import Commit, Repo


class TestProp():
    @property
    def command(self):
        raise NotImplementedError()

    @property
    def exp_name(self):
        raise NotImplementedError()

    @property
    def exp_root(self):
        raise NotImplementedError()

    @property
    def project_name(self):
        raise NotImplementedError()

    @property
    def repo_root(self):
        raise NotImplementedError()

    @property
    def repo_hash(self):
        raise NotImplementedError()

    @property
    def start_time(self):
        raise NotImplementedError()

    @property
    def end_time(self):
        raise NotImplementedError()

    @property
    def end_code(self):
        raise NotImplementedError()

    @property
    def grepo(self) -> Repo:
        raise NotImplementedError()

    @property
    def gcommit(self) -> Commit:
        raise NotImplementedError()

    @property
    def commit_hash(self) -> str:
        raise NotImplementedError()

    @property
    @lru_cache(1)
    def lines(self) -> Dict[str, Dict]:
        raise NotImplementedError()

    @property
    @lru_cache(1)
    def jsons(self):
        raise NotImplementedError()

    @property
    def pkl_keys(self):
        raise NotImplementedError()

    @property
    def uuid(self):
        raise NotImplementedError()

    @property
    def largest_epoch(self):
        raise NotImplementedError()



"""

"""
import numbers
from typing import Any, Iterable

import numpy as np
import torch


class llist(list):
    """
    slice as you like

    Examples:
    >>> res = llist([1,2,3,4])
    ... print(res[0])
    ... print(res[0:3])
    ... print(res[0,2,1])

    >>> idx = torch.randperm(3)
    ... print(res[idx])

    >>> idx = torch.randint(0,3,[3,4])
    ... print(res[idx])

    >>> idx = np.array([1,2,-3])
    ... print(res[idx])

    >>> idx = np.array(2)
    ... print(res[idx])
    """

    def __getitem__(self, i: [int, slice, Iterable]) -> Any:
        if isinstance(i, (slice, numbers.Integral)):
            # numpy.int64 is not an instance of built-in type int
            try:
                res = super().__getitem__(i)
            except IndexError:
                raise IndexError('list index out of range, got {}, but max is {}'.format(i, len(self) - 1))

            if isinstance(res, list):
                return llist(res)
            else:
                return res
        elif isinstance(i, (Iterable)):
            if isinstance(i, torch.Tensor):
                if len(i.shape) == 0:
                    i = i.item()
                    return self.__getitem__(i)
                else:
                    i = i.tolist()
            if isinstance(i, np.ndarray):
                if i.dtype == np.bool:
                    i = np.where(i)[0]

                if len(i.shape) == 0:
                    i = i.tolist()
                    return self.__getitem__(i)
                else:
                    i = i.tolist()

            return llist(self.__getitem__(id) for id in i)



"""

"""

from lumo.base_classes.tree import tree


def test_tree():
    tre = tree()
    tre[1][2][3][4] = 5
    assert tre[1][2][3][4] == 5
    assert isinstance(tre[1][2][3], tree), "should be {}".format(type(tree))
    assert isinstance(tre[1][1][0], tree), "should be {}".format(type(tree))



# copied from https://github.com/briancurtin/deprecation/blob/master/deprecation.py
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import functools
import textwrap
import warnings
from datetime import date
from lumo import __version__

# This is mostly here so automodule docs are ordered more ideally.
__all__ = ["deprecated", "message_location", "fail_if_not_removed",
           "DeprecatedWarning", "UnsupportedWarning"]

#: Location where the details are added to a deprecated docstring
#:
#: When set to ``"bottom"``, the details are appended to the end.
#: When set to ``"top"``, the details are inserted between the
#: summary line and docstring contents.
message_location = "bottom"


class DeprecatedWarning(DeprecationWarning):
    """A warning class for deprecated methods
    This is a specialization of the built-in :class:`DeprecationWarning`,
    adding parameters that allow us to get information into the __str__
    that ends up being sent through the :mod:`warnings` system.
    The attributes aren't able to be retrieved after the warning gets
    raised and passed through the system as only the class--not the
    instance--and message are what gets preserved.
    :param function: The function being deprecated.
    :param deprecated_in: The version that ``function`` is deprecated in
    :param removed_in: The version or :class:`datetime.date` specifying
                       when ``function`` gets removed.
    :param details: Optional details about the deprecation. Most often
                    this will include directions on what to use instead
                    of the now deprecated code.
    """

    def __init__(self, function, deprecated_in, removed_in, details=""):
        # NOTE: The docstring only works for this class if it appears up
        # near the class name, not here inside __init__. I think it has
        # to do with being an exception class.
        self.function = function
        self.deprecated_in = deprecated_in
        self.removed_in = removed_in
        self.details = details
        super(DeprecatedWarning, self).__init__(function, deprecated_in,
                                                removed_in, details)

    def __str__(self):
        # Use a defaultdict to give us the empty string
        # when a part isn't included.
        parts = collections.defaultdict(str)
        parts["function"] = self.function

        if self.deprecated_in:
            parts["deprecated"] = " as of %s" % self.deprecated_in
        if self.removed_in:
            parts["removed"] = " and will be removed {} {}".format("on" if isinstance(self.removed_in, date) else "in",
                                                                   self.removed_in)
        if any([self.deprecated_in, self.removed_in, self.details]):
            parts["period"] = "."
        if self.details:
            parts["details"] = " %s" % self.details

        return ("%(function)s is deprecated%(deprecated)s%(removed)s"
                "%(period)s%(details)s" % (parts))


class UnsupportedWarning(DeprecatedWarning):
    """A warning class for methods to be removed
    This is a subclass of :class:`~deprecation.DeprecatedWarning` and is used
    to output a proper message about a function being unsupported.
    Additionally, the :func:`~deprecation.fail_if_not_removed` decorator
    will handle this warning and cause any tests to fail if the system
    under test uses code that raises this warning.
    """

    def __str__(self):
        parts = collections.defaultdict(str)
        parts["function"] = self.function
        parts["removed"] = self.removed_in

        if self.details:
            parts["details"] = " %s" % self.details

        return ("%(function)s is unsupported as of %(removed)s."
                "%(details)s" % (parts))


def deprecated(deprecated_in=None, removed_in=None,
               details=""):
    """Decorate a function to signify its deprecation
    This function wraps a method that will soon be removed and does two things:
        * The docstring of the method will be modified to include a notice
          about deprecation, e.g., "Deprecated since 0.9.11. Use foo instead."
        * Raises a :class:`~deprecation.DeprecatedWarning`
          via the :mod:`warnings` module, which is a subclass of the built-in
          :class:`DeprecationWarning`. Note that built-in
          :class:`DeprecationWarning`s are ignored by default, so for users
          to be informed of said warnings they will need to enable them--see
          the :mod:`warnings` module documentation for more details.
    :param deprecated_in: The version at which the decorated method is
                          considered deprecated. This will usually be the
                          next version to be released when the decorator is
                          added. The default is **None**, which effectively
                          means immediate deprecation. If this is not
                          specified, then the `removed_in` and
                          `current_version` arguments are ignored.
    :param removed_in: The version or :class:`datetime.date` when the decorated
                       method will be removed. The default is **None**,
                       specifying that the function is not currently planned
                       to be removed.
                       Note: This parameter cannot be set to a value if
                       `deprecated_in=None`.
    :param current_version: The source of version information for the
                            currently running code. This will usually be
                            a `__version__` attribute on your library.
                            The default is `None`.
                            When `current_version=None` the automation to
                            determine if the wrapped function is actually
                            in a period of deprecation or time for removal
                            does not work, causing a
                            :class:`~deprecation.DeprecatedWarning`
                            to be raised in all cases.
    :param details: Extra details to be added to the method docstring and
                    warning. For example, the details may point users to
                    a replacement method, such as "Use the foo_bar
                    method instead". By default there are no details.
    """
    # You can't just jump to removal. It's weird, unfair, and also makes
    # building up the docstring weird.
    if deprecated_in is None and removed_in is not None:
        raise TypeError("Cannot set removed_in to a value "
                        "without also setting deprecated_in")

    current_version = __version__
    # Only warn when it's appropriate. There may be cases when it makes sense
    # to add this decorator before a formal deprecation period begins.
    # In CPython, PendingDeprecatedWarning gets used in that period,
    # so perhaps mimick that at some point.
    is_deprecated = False
    is_unsupported = False

    # StrictVersion won't take a None or a "", so make whatever goes to it
    # is at least *something*. Compare versions only if removed_in is not
    # of type datetime.date
    if isinstance(removed_in, date):
        if date.today() >= removed_in:
            is_unsupported = True
        else:
            is_deprecated = True
    elif current_version:
        current_version = current_version.split('.')

        if (removed_in
                and current_version >= removed_in.split('.')):
            is_unsupported = True
        elif (deprecated_in
              and current_version >= deprecated_in.split('.')):
            is_deprecated = True
    else:
        # If we can't actually calculate that we're in a period of
        # deprecation...well, they used the decorator, so it's deprecated.
        # This will cover the case of someone just using
        # @deprecated("1.0") without the other advantages.
        is_deprecated = True

    should_warn = any([is_deprecated, is_unsupported])

    def _function_wrapper(function):
        if should_warn:
            # Everything *should* have a docstring, but just in case...
            existing_docstring = function.__doc__ or ""

            # The various parts of this decorator being optional makes for
            # a number of ways the deprecation notice could go. The following
            # makes for a nicely constructed sentence with or without any
            # of the parts.

            # If removed_in is a date, use "removed on"
            # If removed_in is a version, use "removed in"
            parts = {
                "deprecated_in":
                    " %s" % deprecated_in if deprecated_in else "",
                "removed_in":
                    "\n   This will be removed {} {}.".format("on" if isinstance(removed_in, date) else "in",
                                                              removed_in) if removed_in else "",
                "details":
                    " %s" % details if details else ""}

            deprecation_note = (".. deprecated::{deprecated_in}"
                                "{removed_in}{details}".format(**parts))

            # default location for insertion of deprecation note
            loc = 1

            # split docstring at first occurrence of newline
            string_list = existing_docstring.split("\n", 1)

            if len(string_list) > 1:
                # With a multi-line docstring, when we modify
                # existing_docstring to add our deprecation_note,
                # if we're not careful we'll interfere with the
                # indentation levels of the contents below the
                # first line, or as PEP 257 calls it, the summary
                # line. Since the summary line can start on the
                # same line as the """, dedenting the whole thing
                # won't help. Split the summary and contents up,
                # dedent the contents independently, then join
                # summary, dedent'ed contents, and our
                # deprecation_note.

                # in-place dedent docstring content
                string_list[1] = textwrap.dedent(string_list[1])

                # we need another newline
                string_list.insert(loc, "\n")

                # change the message_location if we add to end of docstring
                # do this always if not "top"
                if message_location != "top":
                    loc = 3

            # insert deprecation note and dual newline
            string_list.insert(loc, deprecation_note)
            string_list.insert(loc, "\n\n")

            function.__doc__ = "".join(string_list)

        @functools.wraps(function)
        def _inner(*args, **kwargs):
            if should_warn:
                if is_unsupported:
                    cls = UnsupportedWarning
                else:
                    cls = DeprecatedWarning

                the_warning = cls(function.__name__, deprecated_in,
                                  removed_in, details)
                warnings.warn(the_warning, category=DeprecationWarning,
                              stacklevel=2)
            with warnings.catch_warnings(record=True) as caught_warnings:
                warnings.simplefilter("always")
                return function(*args, **kwargs)

        return _inner

    return _function_wrapper


def fail_if_not_removed(method):
    """Decorate a test method to track removal of deprecated code
    This decorator catches :class:`~deprecation.UnsupportedWarning`
    warnings that occur during testing and causes unittests to fail,
    making it easier to keep track of when code should be removed.
    :raises: :class:`AssertionError` if an
             :class:`~deprecation.UnsupportedWarning`
             is raised while running the test method.
    """

    # NOTE(briancurtin): Unless this is named test_inner, nose won't work
    # properly. See Issue #32.
    @functools.wraps(method)
    def test_inner(*args, **kwargs):
        with warnings.catch_warnings(record=True) as caught_warnings:
            warnings.simplefilter("always")
            rv = method(*args, **kwargs)

        for warning in caught_warnings:
            if warning.category == UnsupportedWarning:
                raise AssertionError(
                    ("%s uses a function that should be removed: %s" %
                     (method, str(warning.message))))

        return rv

    return test_inner



""""

"""

from typing import Any




class Merge(type):
    """
    元类，用于将子类和父类共有字典，集合时，子类的覆盖行为改为合并父类的字典，集合

    由于用途特殊，仅识别类变量中以下划线开头的变量
    ::
        class A(metaclass=Merge):
            _dicts = {"1": 2, "3": 4}

        class B(A):
            _dicts = {"5":6,7:8}

        print(B._dicts)

    result:
    >>> {'5': 6, '3': 4, '1': 2, 7: 8}
    """

    def __new__(cls, name, bases, attrs: dict, **kwds):
        for base in bases:
            for key, value in base.__dict__.items():  # type:(str,Any)
                if not key.startswith("_"):
                    continue
                if isinstance(value, set):
                    v = attrs.setdefault(key, set())
                    v.update(value)
                elif isinstance(value, dict):
                    v = attrs.setdefault(key, dict())
                    v.update(value)

        return type.__new__(cls, name, bases, dict(attrs))


class meta_attr(type):
    """用于记录所有attr子类的类名和类信息，用于序列化和反序列化"""

    def __new__(cls, *args: Any, **kwargs: Any):
        from lumo.base_classes.attr import _attr_clss
        cls = type.__new__(cls, *args, **kwargs)
        _attr_clss[cls.__name__] = cls
        return cls






import os
from torch import distributed as dist


def local_rank():
    rank = os.environ.get('LOCAL_RANK', -1)
    if rank == -1:
        if dist.is_available() and dist.is_initialized():
            rank = dist.get_rank()

    return int(rank)


def world_size():
    size = os.environ.get('WORLD_SIZE', 0)
    if size == 0:
        if dist.is_available() and dist.is_initialized():
            size = dist.get_world_size()
    return int(size)


def is_dist():
    return local_rank() >= 0



import numpy as np
from scipy.io import wavfile


def add_wgn(s, var=1e-4):
    """
        Add white Gaussian noise to signal
        If no variance is given, simply add jitter. 
        Jitter helps eliminate all-zero values.
        """
    np.random.seed(0)
    noise = np.random.normal(0, var, len(s))
    return s + noise


def read_wav(filename):
    """
        read wav file.
        Normalizes signal to values between -1 and 1.
        Also add some jitter to remove all-zero segments."""
    fs, s = wavfile.read(filename)  # scipy reads int
    s = np.array(s)
    ns = add_wgn(s)  # Add jitter for numerical stability
    return fs, s, ns


def enframe(x, win_len, hop_len):
    """
        receives a 1D numpy array and divides it into frames.
        outputs a numpy matrix with the frames on the rows.
    """
    x = np.squeeze(x)
    if x.ndim != 1:
        raise TypeError("enframe input must be a 1-dimensional array.")
    n_frames = 1 + np.int(np.floor((len(x) - win_len) / float(hop_len)))
    x_framed = np.zeros((n_frames, win_len))
    for i in range(n_frames):
        x_framed[i] = x[i * hop_len: i * hop_len + win_len]
    return x_framed


def deframe(x_framed, win_len, hop_len):
    '''
        interpolates 1D data with framed alignments into persample values.
        This function helps as a visual aid and can also be used to change 
        frame-rate for features, e.g. energy, zero-crossing, etc.
        '''
    n_frames = len(x_framed)
    n_samples = n_frames * hop_len + win_len
    x_samples = np.zeros((n_samples, 1))
    for i in range(n_frames):
        x_samples[i * hop_len: i * hop_len + win_len] = x_framed[i]
    return x_samples


if __name__ == '__main__':
    pass



from typing import Union, Optional, Sequence, Mapping

import torch
from torch import nn

from torch.utils.data import DataLoader

from lumo import Trainer, DataModule, callbacks, Params, Meter
from lumo.base_classes import TrainerStage
from lumo.kit import ParamsType, DataModuleMix


class TempCallback(callbacks.TrainCallback):

    def __init__(self) -> None:
        super().__init__()
        self.c = 0

    def on_train_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        self.c += 1

    def on_train_epoch_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        self.c += 1

    def on_test_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        self.c += 1

    def on_eval_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        self.c += 1

    def on_train_step_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        self.c += 1

    def on_eval_step_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        self.c += 1

    def on_test_step_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        self.c += 1

    def on_train_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        self.c += 1

    def on_train_epoch_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        self.c += 1

    def on_test_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        self.c += 1

    def on_eval_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        self.c += 1

    def on_train_step_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        self.c += 1

    def on_eval_step_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        self.c += 1

    def on_test_step_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        self.c += 1

    def on_predict_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        self.c += 1

    def on_predict_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        self.c += 1

    def on_inference_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        self.c += 1

    def on_inference_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        self.c += 1


class CBTrainer(Trainer):
    def stop_train(self):
        super().stop_train()

    def stop_train_epoch(self):
        super().stop_train_epoch()

    def to_stage(self, stage: TrainerStage):
        super().to_stage(stage)

    def to_device(self, item: Optional[Union[nn.Module, torch.Tensor, Sequence, Mapping]] = None,
                  device_args_kwargs=None):
        return super().to_device(item, device_args_kwargs)

    def prepare_dataloader(self, stage: TrainerStage, dataloader=None):
        return super().prepare_dataloader(stage, dataloader)

    def train(self, dataloader: Union[DataLoader, DataModuleMix] = None):
        return super().train(dataloader)

    def train_epoch(self, dataloader: DataLoader):
        return super().train_epoch(dataloader)

    def train_step(self, idx, batch, params: ParamsType, *args, **kwargs):
        super().train_step(idx, batch, params, *args, **kwargs)

    def evaluate(self, dataloader: Union[DataLoader, DataModule] = None):
        return super().evaluate(dataloader)

    def evaluate_step(self, idx, batch, params: ParamsType, *args, **kwargs):
        super().evaluate_step(idx, batch, params, *args, **kwargs)

    def test(self, dataloader: Union[DataLoader, DataModule] = None):
        return super().test(dataloader)

    def test_step(self, idx, batch, params: ParamsType, *args, **kwargs):
        super().test_step(idx, batch, params, *args, **kwargs)

    def inference(self, batch):
        super().inference(batch)

    def predict(self, batch):
        super().predict(batch)

    def ioptims(self, params: ParamsType):
        super().ioptims(params)

    def icallbacks(self, params: ParamsType):
        super().icallbacks(params)

    def imodels(self, params: ParamsType):
        super().imodels(params)


def test_callback():
    params = Params()




from lumo import Saver

saver = Saver('./sav')
saver.save_checkpoint(1, {1, 2, 3})
saver.save_checkpoint(2, {1, 2, 3})
saver.save_checkpoint(3, {1, 2, 3})
saver.save_checkpoint(4, {1, 2, 3}, is_best=True)
saver.save_keypoint(3, {1, 2, 3})
saver.save_keypoint(3, {1, 2, 3})
fn = saver.save_model(4, {1, 2, 3}, is_best=True)

print(fn)

print(saver.load_state_dict(fn))

# saver.load_state_dict()
print(saver.list_models())
print(saver.list_checkpoints())
print(saver.list_keypoints())




import hashlib
import os
from itertools import chain

from torch import nn

from . import re


def file_atime_hash(file):
    """
    calculate hash of given file's atime
    atime : time of last access
    """
    return string_hash(str(os.path.getatime(file)))


def string_hash(*str):
    """calculate hash of given string list"""
    hl = hashlib.md5()
    for s in str:
        hl.update(s.encode(encoding='utf-8'))
    return hl.hexdigest()[:16]


def file_hash(file: str) -> str:
    """calculate hash of given file path"""
    hl = hashlib.md5()
    with open(file, 'rb') as r:
        hl.update(r.read())
    return hl.hexdigest()[:16]


def filter_filename(title: str, substr='-'):
    """replace invalid string of given file path by `substr`"""
    title = re.sub('[\/:*?"<>|]', substr, title)  # 去掉非法字符
    return title


def inthash(value) -> int:
    return int(hash(value), 16)


def hash(value, hexint=False) -> str:
    """try to calculate hash of any given object"""
    import hashlib
    from collections.abc import Iterable
    from numbers import Number
    import numpy as np
    from torch import Tensor
    hl = hashlib.md5()

    if isinstance(value, (np.ndarray, Tensor)):
        if isinstance(value, Tensor):
            value = value.detach_().cpu().numpy()
        try:
            value = value.item()
        except ValueError:  # not an one element tensor
            value = str(value)
    if isinstance(value, (Number)):
        value = str(value)

    if isinstance(value, dict):
        for k in sorted(value.keys()):
            v = value[k]
            hl.update(str(k).encode(encoding='utf-8'))
            hl.update(hash(v).encode(encoding='utf-8'))
    elif isinstance(value, str):
        hl.update(value.encode(encoding='utf-8'))
    elif isinstance(value, Iterable):
        for v in value:
            hl.update(hash(v).encode(encoding='utf-8'))

    res = hl.hexdigest()
    if hexint:
        res = str(int(res, 16))
    return res


def hash_model(model: nn.Module):
    """
    If two model have equal model hash, then their paramters can be shared with each other.
    Commonly used for loading pretrained model.
    """
    hl = hashlib.md5()

    named_iter = chain(
        model.named_modules(),
        model.named_parameters(),
        model.named_children(),
        model.named_buffers(),
    )
    for k, _ in named_iter:  # type:str
        hl.update(k.encode(encoding='utf-8'))

    return hl.hexdigest()



from lumo.kit.builder import DatasetBuilder


def create_dataset_builder():
    builder = (
        DatasetBuilder()
            .add_input(name='xs', source=range(1000))
            .add_input(name='ys', source=range(1, 1001))
            .add_output(name='xs', outkey='xs1')
            .add_output(name='xs', outkey='xs2')
            .add_output(name='ys', outkey='ys1')
            .add_output_transform('xs1', lambda x: x + 1)
            .add_output_transform('ys1', lambda x: x - 1)
    )
    return builder


def test_builder_base():
    builder = create_dataset_builder()

    try:
        builder.add_input('xs', source=range(1000))
        assert False
    except:
        pass
    try:
        builder.add_input('xxs', source=range(1001))
        assert False
    except:
        pass

    assert len(builder) == 1000

    sub_builder = builder.subset(range(500), copy=True)
    assert len(builder) == 1000 and len(sub_builder) == 500

    assert sub_builder[499]['xs1'] == 500
    assert sub_builder[499]['ys1'] == 499
    assert sub_builder[499]['xs2'] == 499



"""
for calculating and recording accuracy.
"""
import torch

from lumo import Meter, Params
try:
    import sklearn
except ImportError:
    import warnings
    warnings.warn("You need to install scikit-learn to use UnsupervisedAccMixin class ")

class AccMixin():
    pass


class UnsupervisedAccMixin(AccMixin):
    def acc_assigned_(self):
        pass  # TODO

    def acc_nmi_(self, pred_labels: torch.Tensor, labels: torch.Tensor,
                 meter: Meter = None, name='nmi'):
        """
        Normalized Mutual Information for Unsupervised
        Args:
            pred_labels: (batchsize, )
            labels: (batchsize, )
            meter:
            name:

        Returns:
            nmi
        """
        from sklearn import metrics
        y_true = labels.cpu().numpy()
        y_pred = pred_labels.cpu().numpy()

        acc = metrics.normalized_mutual_info_score(y_true, y_pred, average_method="arithmetic")
        if meter is not None:
            meter[name] = acc
        return acc

    def acc_ari_(self, pred_labels: torch.Tensor, labels: torch.Tensor,
                 meter: Meter = None, name='ari'):
        """
        Rand index adjusted for Unsupervised
        Args:
            pred_labels: (batchsize, )
            labels: (batchsize, )
            meter:
            name:

        Returns:
            nmi
        """
        from sklearn import metrics

        y_true = labels.cpu().numpy()
        y_pred = pred_labels.cpu().numpy()

        acc = metrics.adjusted_rand_score(y_true, y_pred)
        if meter is not None:
            meter[name] = acc
        return acc

    def test_eval_logic(self, dataloader, param: Params):
        from lumo.calculate import accuracy as acc
        from sklearn import metrics
        import numpy as np
        with torch.no_grad():
            meter = Meter()
            y_trues = []
            y_preds = []
            for xs, y_true in dataloader:
                preds = self.predict(xs)  # type:torch.Tensor

                y_pred = preds.argmax(dim=-1).cpu().numpy()
                y_true = y_true.cpu().numpy()

                y_preds.extend(y_pred)
                y_trues.extend(y_true)

            meter.nmi = metrics.normalized_mutual_info_score(
                np.array(y_trues), np.array(y_preds),
                average_method="arithmetic")

            meter.ari = metrics.adjusted_rand_score(np.array(y_trues), np.array(y_preds))

        return meter


class ClassifyAccMixin(AccMixin):
    def acc_precise_(self, pred_labels: torch.Tensor, labels, meter: Meter = None, name='acc'):
        """train batch accuracy"""
        with torch.no_grad():
            maxid = pred_labels
            total = labels.shape[0]
            top1 = (labels == maxid).sum().float() / total
            top1 = top1.item()

        if meter is not None:
            meter[name] = top1
            meter.percent(name)
        return top1

    def test_eval_logic(self, dataloader, param: Params):
        from lumo.calculate import accuracy as acc
        param.topk = param.default([1, 5])
        with torch.no_grad():
            count_dict = Meter()
            for xs, labels in dataloader:
                preds = self.predict(xs)
                total, topk_res = acc.classify(preds, labels, topk=param.topk)
                count_dict["total"] += total
                for i, topi_res in zip(param.topk, topk_res):
                    count_dict["top{}".format(i)] += topi_res
        return count_dict



"""
for calculating and recording loss.
"""
from typing import List

from torch.nn import functional as F

import torch
from lumo import Meter


class Loss():
    def to_mid(self, xs) -> torch.Tensor:
        """input2mid"""
        raise NotImplementedError()

    def to_logits(self, xs) -> torch.Tensor:
        """input2logits"""
        raise NotImplementedError()

    def mid_to_logits(self, xs) -> torch.Tensor:
        raise NotImplementedError()


class MSELoss(Loss):
    def loss_mse_(self, logits: torch.Tensor, targets: torch.Tensor,
                  w_mse=1,
                  meter: Meter = None, name: str = 'Lmse'):
        loss = torch.mean((F.softmax(logits, dim=1) - targets) ** 2) * w_mse
        if meter is not None:
            meter[name] = loss
        return loss

    def loss_mse_with_labels_(self, logits: torch.Tensor, labels: torch.Tensor,
                              meter: Meter, name: str = 'Lmse', w_mse=1):
        from lumo.calculate.tensor import onehot
        targets = onehot(labels, logits.shape[-1])
        meter[name] = torch.mean((F.softmax(logits, dim=1) - targets) ** 2) * w_mse
        return meter[name]


class CELoss(Loss):
    def loss_ce_(self, logits: torch.Tensor, labels: torch.Tensor,
                 w_ce=1,
                 meter: Meter = None, name: str = 'Lce'):
        loss = F.cross_entropy(logits, labels) * w_ce
        if meter is not None:
            meter[name] = loss
        return loss

    def loss_ce_with_targets_(self,
                              logits: torch.Tensor, targets: torch.Tensor,
                              w_ce=1,
                              meter: Meter = None, name: str = 'Lce'):
        loss = -torch.mean(torch.sum(F.log_softmax(logits, dim=1) * targets, dim=1)) * w_ce
        if meter is not None:
            meter[name] = loss
        return loss


class MinENTLoss(Loss):
    def loss_minent_(self, logits, w_ent=1,
                     meter: Meter = None, name: str = 'Lent'):
        loss = - torch.sum(F.log_softmax(logits, dim=-1) * F.softmax(logits, dim=-1), dim=-1).mean() * w_ent
        if meter is not None:
            meter[name] = loss
        return loss


class L2Loss(Loss):
    def loss_l2_reg_(self, tensors: List[torch.Tensor], w_l2=1,
                     meter: Meter = None, name: str = 'Lreg'):
        loss = sum([(tensor ** 2).sum(dim=-1).mean() for tensor in tensors]) * w_l2
        if meter is not None:
            meter[name] = loss
        return loss


# class TripletLoss(Loss):
# def loss_l2_reg_(self, tensors: List[torch.Tensor], w_l2=1,
#                  meter: Meter = None, name: str = 'Lreg'):
#     loss = sum([(tensor ** 2).sum(dim=-1).mean() for tensor in tensors]) * w_l2
#     if meter is not None:
#         meter[name] = loss
#     return loss


class SimCLRLoss(Loss):
    def loss_sim_(self, features: torch.Tensor,
                  temperature=0.5,
                  meter: Meter = None,
                  name: str = 'Lsim'):
        """

        :param features: [batchsize, 2, feature_dim]
        :param temperature:
        :param meter:
        :param name:
        :return:
        """
        b, n, dim = features.size()
        assert (n == 2)
        mask = torch.eye(b, dtype=torch.float32).cuda()

        contrast_features = torch.cat(torch.unbind(features, dim=1), dim=0)
        anchor = features[:, 0]

        # Dot product
        dot_product = torch.matmul(anchor, contrast_features.T) / temperature  # # 两两之间的相似度

        # Log-sum trick for numerical stability
        logits_max, _ = torch.max(dot_product, dim=1, keepdim=True)
        logits = dot_product - logits_max.detach()  # 两两相似度 - 第一次增广的自身的相似度

        mask = mask.repeat(1, 2)

        # 排除第一次增广自身相似度的 mask，# 也即分母部份 logits 的 mask
        logits_mask = torch.scatter(torch.ones_like(mask), 1, torch.arange(b).view(-1, 1).cuda(), 0)
        mask = mask * logits_mask  # 第一次样本增广和第二次样本增广的相似度的 mask

        # Log-softmax
        exp_logits = torch.exp(logits) * logits_mask  # 分母部份的底

        # logits 的 exp 和 log 抵消了，所以只要logits本身即可，而 第二项则是 log 内的分母
        log_prob = logits - torch.log(exp_logits.sum(1, keepdim=True))

        # Mean log-likelihood for positive
        # mask * log_prob 的 每行都表示 分子为正例，分母为全部的一个 lij，最后求和得到所有 lij 的损失
        loss = - ((mask * log_prob).sum(1) / mask.sum(1)).mean()

        if meter is not None:
            meter[name] = loss

        return loss


class KLLoss(Loss):
    def loss_kl_(self, input_logits: torch.Tensor, target_logits: torch.Tensor,
                 meter: Meter = None, name: str = 'Lcon'):
        """Takes softmax on both sides and returns KL divergence
        from ICT implement
        Note:
        - Returns the sum over all examples. Divide by the batch size afterwards
          if you want the mean.
        - Sends gradients to inputs but not the targets.
        """
        assert input_logits.size() == target_logits.size()
        input_log_softmax = F.log_softmax(input_logits, dim=-1)
        target_softmax = F.softmax(target_logits, dim=1)
        loss = F.kl_div(input_log_softmax, target_softmax)

        if meter is not None:
            meter[name] = loss

        return loss


class ContranstiveLoss(Loss):
    def loss_contrastive_(self, queries: torch.Tensor, keys: torch.Tensor, temperature=0.1,
                          meter: Meter = None, name: str = 'Lcs'):
        """
        loss_contrastive_(torch.rand(4, 128), torch.rand(4, 128))

        :param queries:
        :param keys:
        :param temperature:
        :return:
        """
        b, device = queries.shape[0], queries.device
        logits = queries @ keys.t()
        logits = logits - logits.max(dim=-1, keepdim=True).values
        logits /= temperature
        loss = F.cross_entropy(logits, torch.arange(b, device=device))
        if meter is not None:
            meter[name] = loss

        return loss



import os
import shutil
import textwrap
from collections import namedtuple
from typing import Optional, Union, List

from lumo.utils import safe_io as io

state_dict_tuple = namedtuple('state_dict_tuplt', ['state_dict', 'meta_info'], defaults=[None])


class Saver:
    """
    Write state_dict into test dirs, record save log into <repo working dir>/.lumo/save.<exp_name>.log

    format:
        <test_name> -> <fn>

    Raises:
        When save/load operations happend, you may meet Out Of Space, FileNotExist, or other problems,
        then Exception or its stack infomation will be printed in stderr (not be raised), and `False`/`None` value will be returned by
        save_xx and load_xx functions.
        You can handle these values and stop your program manully, or just let it go.
    """

    def __init__(self, save_dir):
        self._save_dir = save_dir

    @property
    def save_dir(self):
        os.makedirs(self._save_dir, exist_ok=True)
        return self._save_dir

    def _guess_abs_path(self, fn: str) -> str:
        path = fn
        if os.path.basename(fn) == fn:
            path = os.path.join(self.save_dir, fn)
        return path

    def _create_state_dict_name(self, step=0, replacement: bool = True,
                                prefix=None, ignore_number=False, suffix='pt') -> str:
        """
        create path for saving state_dict
        """
        offset = 0

        def _create_path():
            res = []
            res.append(prefix)
            if not replacement:
                res.append(f"{offset:01d}")
            if ignore_number:
                res.append(None)
            else:
                res.append(f"{step:06d}")
            res.append(suffix)
            fn = '.'.join(list(filter(lambda x: x is not None, res)))
            return os.path.join(self.save_dir, fn)

        path = _create_path()
        if not replacement:
            while os.path.exists(path):
                offset += 1
                path = _create_path()
        return path

    def dump_state_dict(self, obj, fn, meta_info: Union[str, dict] = None):
        """

        Args:
            obj: picklable object
            fn: path-like object
            meta_info: meta infomation for the dumped object.

        Returns:
            saved filepath, None if something went wrong.
        """
        res = io.dump_state_dict(obj, fn)
        if res and meta_info is not None:
            if isinstance(meta_info, str):
                meta_info = {'msg': meta_info}
            json_fn = f"{fn}.json"
            io.dump_json(meta_info, json_fn)
        return res

    def load_state_dict(self, fn: str, with_meta=False, map_location='cpu') -> state_dict_tuple:
        """

        Args:
            fn: pickled file path.
            with_meta: whether to get its meta infomation at the same time, default is False.
            map_location:

        Returns:
            If `with_meta=True`, a `state_dict_tuple` object will be returned, you can unpack this tuple directly:
            ```
            res = load_state_dict(...)
            ckpt, info = res
            ```
            or `ckpt` object will be returned alone.

            if something went wrong, `ckpt` or `info` will be a None object.
        """
        path = self._guess_abs_path(fn)
        ckpt = io.load_state_dict(path, map_location=map_location)

        if not with_meta:
            return ckpt

        info = self.load_meta_info(fn)
        return state_dict_tuple(ckpt, info)

    def load_meta_info(self, fn: str) -> dict:
        """
        load
        Args:
            fn: pickled file path, not the metainfo file path(which ends with suffix '.json').

        Returns:
            info object, or None if something went wrong.

        """
        path = self._guess_abs_path(fn)
        info_path = f"{path}.json"
        info = io.load_json(info_path)
        return info

    def save_keypoint(self, steps: int, state_dict, meta_info: Union[str, dict] = None) -> str:
        """
        save a object which is keypoint.
        Args:
            steps:
            state_dict:
            meta_info:

        Returns:

        """
        path = self._create_state_dict_name(steps, replacement=True, prefix='key')
        self.dump_state_dict(state_dict, path, meta_info=meta_info)
        return path

    def save_checkpoint(self, step: int, state_dict, meta_info: Union[str, dict] = None,
                        max_keep=10, is_best=False) -> Optional[str]:
        """
        save a object as a deleteable

        Args:
            step: progress indecator, using global_steps or epoch count is recommanded.
            state_dict: state dict
            meta_info: meta information for this state dict
            max_keep:
            is_best: Whether this checkpoint is currently best. If True, the best checkpoint
                file will be replaced by this one.

        Returns:
            Saved checpoint path, or None if something is wrong.

        """
        path = self._create_state_dict_name(step, replacement=False, prefix='checkpoints')
        res = self.dump_state_dict(state_dict, path, meta_info=meta_info)
        del state_dict
        history = self.list_checkpoints()
        if len(history) > max_keep:
            [os.remove(os.path.join(self.save_dir, i)) for i in history[:-max_keep]]

        if res:
            if is_best:
                best_path = self._create_state_dict_name(step,
                                                         replacement=True,
                                                         prefix='best.checkpoint',
                                                         ignore_number=True)
                shutil.copy2(path, best_path)
                src_json_fn = f"{path}.json"
                if os.path.exists(src_json_fn):
                    tgt_json_fn = f"{best_path}.json"
                    shutil.copy2(src_json_fn, tgt_json_fn)

            return path
        else:
            return None

    def save_model(self, step: int, state_dict, meta_info: Union[str, dict] = None, is_best=False) -> str:
        """
        save model

        Args:
            steps: progress indecator, using global_steps or epoch count is recommanded.
            state_dict: state dict
            meta_info: meta information for this state dict
            is_best: Whether this checkpoint is currently best. If True, the best checkpoint
                file will be replaced by this one.

        Returns:
            Saved checpoint path, or None if something is wrong.

        """
        path = self._create_state_dict_name(step, replacement=True, prefix='model')
        res = self.dump_state_dict(state_dict, path, meta_info=meta_info)
        del state_dict
        if res:
            if is_best:
                best_path = self._create_state_dict_name(step,
                                                         replacement=True,
                                                         prefix='best.model',
                                                         ignore_number=True)
                shutil.copy2(path, best_path)
                src_json_fn = f"{path}.json"
                if os.path.exists(src_json_fn):
                    tgt_json_fn = f"{best_path}.json"
                    shutil.copy2(src_json_fn, tgt_json_fn)
            return path
        else:
            return None

    def load_checkpoint(self, index=-1, best_if_exist=False, fn=None, with_meta=False, map_location='cpu'):
        if fn is None and best_if_exist:
            fn = self.best_checkpoint()
        if fn is None:
            try:
                fn = self.list_checkpoints()[index]
            except:
                pass
        if fn is not None:
            return self.load_state_dict(fn, with_meta, map_location)
        return None

    def load_keypoint(self, index=-1, fn=None, with_meta=False, map_location='cpu'):
        if fn is None:
            try:
                fn = self.list_keypoints()[index]
            except:
                pass
        if fn is not None:
            return self.load_state_dict(fn, with_meta, map_location)
        return None

    def load_model(self, index=-1, best_if_exist=False, fn=None, with_meta=False, map_location='cpu'):
        if fn is None and best_if_exist:
            fn = self.best_model()
        if fn is None:
            try:
                fn = self.list_models()[index]
            except:
                pass
        if fn is not None:
            return self.load_state_dict(fn, with_meta, map_location)
        return None

    def best_checkpoint(self):
        fn = os.path.join(self.save_dir, 'best.checkpoint.pt')
        if os.path.exists(fn):
            return fn
        return None

    def best_model(self):
        fn = os.path.join(self.save_dir, 'best.model.pt')
        if os.path.exists(fn):
            return fn
        return None

    def _is_pkl(self, x: str, start, end):
        return x.startswith(start) and x.endswith(end)

    def list_checkpoints(self) -> List[str]:
        return sorted(list(filter(lambda x: self._is_pkl(x, 'checkpoints', 'pt'),
                                  os.listdir(self.save_dir))),
                      key=lambda x: os.stat(os.path.join(self.save_dir, x)).st_ctime)

    def list_keypoints(self) -> List[str]:
        return sorted(list(filter(lambda x: self._is_pkl(x, 'key', 'pt'),
                                  os.listdir(self.save_dir))),
                      key=lambda x: os.stat(os.path.join(self.save_dir, x)).st_ctime)

    def list_models(self) -> List[str]:
        return sorted(list(filter(lambda x: self._is_pkl(x, 'model', 'pt'),
                                  os.listdir(self.save_dir))),
                      key=lambda x: os.stat(os.path.join(self.save_dir, x)).st_ctime)

    def list(self):
        return sorted(os.listdir(self.save_dir))

    def summary(self):
        print(f"saved in : {self.save_dir}")
        print(textwrap.indent('\n'.join(self.list()), ' - '))



"""

"""

import os
import sys
from functools import wraps
import inspect

from .mixin import CallbackMix
from .meter import AvgMeter
from .meter import Meter
from .params import Params
from ..base_classes.trickitems import NoneItem, AvgItem
from ..calculate.schedule import Schedule, ScheduleList
from ..utils.timing import format_second
from typing import TYPE_CHECKING

func_map = {
    'evaluate': 'eval',
    'evaluate_step': 'eval_step'
}

from .trainer import Trainer, TrainerResult


# if TYPE_CHECKING:


def map_func_name(name):
    return func_map.get(name, name)


class BaseCallback():
    """
    base callback class

    only have two methods `on_begin()` and `on_end()`.

    for simpler using, see TrainCallback.
    """
    priority = 0  # type:int # All callbacks in lumo will have priority in range 0-100
    only_single_gpu = False  # only hooked in single gpu mode
    only_main_process = False  # whether can be hooked in children process( local_rank > 0)

    def __new__(cls, *_, **__):
        self = super().__new__(cls)
        self._hooked = None

        def ecp_wrap(func):
            """同一个异常第一次调用的时候运行"""

            @wraps(func)
            def on_exception(hooked: Trainer, tfunc, params: Params, e: BaseException, *args, **kwargs):
                self.ecp = getattr(self, "ecp", None)
                res = None
                if self.ecp != e:
                    res = self.on_first_exception(hooked, tfunc, params, e, *args, **kwargs)
                    self.ecp = e

                eres = func(hooked, tfunc, params, e, *args, **kwargs)
                if res is None:
                    return eres
                else:
                    return res

            return on_exception

        self.on_exception = ecp_wrap(self.on_exception)
        return self

    def getfuncargs(self, func, *args, **kwargs):
        res = inspect.getcallargs(func, *args, **kwargs)
        if 'self' in res:
            res.pop('self')
        return res

    def on_hooked(self, source: Trainer, params: Params):
        """called when callback hooked trainer"""
        pass

    def on_first_exception(self, source: Trainer, func, params: Params, e: BaseException, *args, **kwargs):
        """
        when an exception was raised in some function, on_exception() will be called.

        如果异常发生在一个嵌套调用的函数中，那么该异常会在每一层函数都raise一次。

        该方法将被调用当该异常第一次raise出来的时候。
        该方法在 __new__ 中作了处理逻辑，不受继承关系影响
        """
        pass

    def on_exception(self, source: Trainer, func, params: Params, e: BaseException, *args, **kwargs):
        """called when exception raised in some function"""
        return False

    def on_hook_failed(self, source, message):
        """Any reason when callback cannot hook on trainer"""
        pass

    def on_begin(self, source: Trainer, func, params: Params, *args, **kwargs):
        """called before trainer.func is called"""
        cb_func = getattr(self, f"on_{map_func_name(func.__name__)}_begin", None)
        if cb_func is not None:
            cb_func(source, func, params, *args, **kwargs)

    def on_end(self, source: Trainer, func, params: Params, result, *args, **kwargs):
        cb_func = getattr(self, f"on_{map_func_name(func.__name__)}_end", None)
        if cb_func is not None:
            cb_func(source, func, params, result, *args, **kwargs)

    def __le__(self, other):
        return self.priority <= other.priority

    def __lt__(self, other):
        return self.priority < other.priority

    def hook(self, source: CallbackMix):
        source.reload_callback(self)

    def unhook(self):
        self._hooked.remove_callback(self)

    def _repr_by_val(self, *vals):
        vstr = "; ".join(["{}={}".format(val, str(getattr(self, val, None))) for val in vals])
        return "<{}([{}]) at 0x{:016X}>".format(self.__class__.__name__, vstr, id(self))

    def __repr__(self) -> str:
        return self._repr_by_val("priority")


class TrainCallback(BaseCallback):
    """
    实现了一般训练过程中的函数函数回调，主要将 on_begin() / on_end() 方法分发到各具体的回调方法中
    """

    def on_train_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_train_epoch_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_test_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_eval_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_train_step_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_eval_step_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_test_step_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_train_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        pass

    def on_train_epoch_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        pass

    def on_test_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        pass

    def on_eval_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        pass

    def on_train_step_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass

    def on_eval_step_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass

    def on_test_step_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        pass

    def on_predict_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_predict_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass

    def on_inference_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_inference_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass


class SaveLoadCallback(BaseCallback):
    def on_save_keypoint_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_save_keypoint_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass

    def on_save_model_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_save_model_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass

    def on_save_checkpoint_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_save_checkpoint_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass

    def on_load_state_dict_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_load_state_dict_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass


class InitialCallback(BaseCallback):
    def on_ioptims_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_ioptims_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass

    def on_imodels_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_imodels_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass

    def on_prepare_dataloader_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        pass

    def on_prepare_dataloader_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        pass


class EvalCallback(TrainCallback):
    """
    """
    only_main_process = True

    def __init__(self, eval_per_epoch=1, test_per_epoch=10):
        self.eval_in_per_epoch = eval_per_epoch
        self.test_in_per_epoch = test_per_epoch

        # evaluate/test on train end
        self._last_eval = -1
        self._last_test = -1

    def _test_or_eval(self, params: Params, trainer: Trainer):
        if self.eval_in_per_epoch is not None and self.eval_in_per_epoch > 0:
            if params.eidx % self.eval_in_per_epoch == self.eval_in_per_epoch - 1:
                self._last_eval = params.eidx
                trainer.evaluate()
        if self.test_in_per_epoch is not None and self.test_in_per_epoch > 0:
            if params.eidx % self.test_in_per_epoch == self.test_in_per_epoch - 1:
                self._last_test = params.eidx
                trainer.test()

    def on_train_epoch_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        self._test_or_eval(params, trainer)

    def on_train_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        if self._last_eval != params.eidx:
            trainer.evaluate()
        if self._last_test != params.eidx:
            trainer.test()

    def __repr__(self):
        return self._repr_by_val("eval_in_per_epoch", "test_per_epoch")


class LoggerCallback(TrainCallback, InitialCallback, SaveLoadCallback):
    """
    用于日志输出的回调，当 BaseTrainer 在 epoch / batch 等级别的训练结束、异常发生等过程后，Logger 会对这些事件，
    或方法返回的结果进行输出。

    一般情况下 Logger 支持所有类型输出，但如果使用 Meter 类进行包装，会有更好的输出形式
    """
    only_main_process = True
    priority = 100

    def __init__(self, avg=True, step_frequence=3, breakline_in=1000):
        self.avg = avg
        self.step_frequence = step_frequence
        self.breakline = breakline_in

    def on_hooked(self, source: Trainer, params: Params):
        source.logger.raw(' '.join(sys.argv))
        source.logger.info("Exp BaseDir", os.path.abspath(source.exp.exp_root))
        source.logger.info("Exp Trainer", source.__class__.__name__)
        source.logger.info("Exp Params")
        source.logger.raw(params)
        self.start = 0
        self.cur = None
        self._meter = None

    def on_train_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        from ..utils.timing import TimeIt

        self.start = params.eidx
        self.traintime = TimeIt()
        self.traintime.start()
        trainer.logger.info('[[Train]]')
        super().on_train_begin(trainer, func, params, *args, **kwargs)

    def on_train_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        self.traintime.end()
        meter = result.meter
        if meter is None:
            meter = ""
        trainer.logger.info(f"[[Train End in {format_second(self.traintime['use'])}]]", meter)

    @property
    def meter(self):
        if self._meter is None:
            self.reset_meter()
        return self._meter

    def on_train_epoch_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        from ..utils.timing import TimeIt
        self.reset_meter()
        self.epochtime = TimeIt()
        self.epochtime.start()
        trainer.logger.info("{}/{}".format(params.eidx, params.epoch))

    def on_train_epoch_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        self.traintime.mark("epoch")
        self.epochtime.end()
        if self.cur is None:
            self.cur = params.eidx

        avg = self.traintime["use"] / (self.cur - self.start + 1)
        self.cur += 1
        last = (params.epoch - params.eidx) * avg

        tm = Meter()
        tm.train = format_second(self.traintime["use"])
        tm.epoch = format_second(self.epochtime["use"])
        tm.avg = format_second(avg)
        tm.last = format_second(last)
        trainer.logger.info(tm)

    def reset_meter(self):
        if self.avg:
            meter = AvgMeter()
        else:
            meter = Meter()
        self._meter = meter

    def _need_log(self, step):
        return self.step_frequence > 0 and step % self.step_frequence == 0

    def _need_breakline(self, step):
        return self.breakline > 0 and step % self.breakline == 0

    def on_train_step_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        meter = trainer._wrap_result(meter)
        self.meter.update(meter)
        meter = self.meter
        if self._need_breakline(params.idx):
            trainer.logger.newline()
            self.reset_meter()
        if self._need_log(params.idx):
            trainer.logger.inline("{}/{}".format(params.idx + 1, len(trainer.train_dataloader)), meter, fix=1)

    def on_first_exception(self, source: Trainer, func, params: Params, e: BaseException, *args, **kwargs):
        source.logger.error("{} raised".format(e.__class__.__name__))

    def on_test_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        self.reset_meter()
        trainer.logger.info("[[Test]]")

    def on_test_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        meter = result.meter
        if meter is None:
            meter = ""
        trainer.logger.info("[[Test End]]", meter)

    def on_test_step_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        meter = trainer._wrap_result(meter)
        self.meter.update(meter)
        meter = self.meter
        trainer.logger.inline("{}/{}".format(params.idx + 1, len(trainer.test_dataloader)), meter, fix=1)

    def on_eval_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        self.reset_meter()
        trainer.logger.info("[[Eval]]")

    def on_eval_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        meter = result.meter
        if meter is None:
            meter = ""
        trainer.logger.info("[[Eval End]]", meter)

    def on_eval_step_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        meter = trainer._wrap_result(meter)
        self.meter.update(meter)
        meter = self.meter
        trainer.logger.inline("{}/{}".format(params.idx + 1, len(trainer.val_dataloader)), meter, fix=1)

    def on_save_checkpoint_end(self, trainer: Trainer, func, params: Params, result: str, *args, **kwargs):
        trainer.logger.info(f'Saved checkpoint in {result}')

    def on_save_model_end(self, trainer: Trainer, func, params: Params, result: str, *args, **kwargs):
        trainer.logger.info(f'Saved mdoel in {result}')

    def on_save_keypoint_end(self, trainer: Trainer, func, params: Params, result: str, *args, **kwargs):
        trainer.logger.info(f'Saved keypoint in {result}')

    def on_load_state_dict_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        res = self.getfuncargs(func, *args, **kwargs)
        object = res['object']
        if isinstance(object, str):
            trainer.logger.info(f'Load state dict from {object}')

        if meter is not None:
            trainer.logger.info(f'Loaded with meta info:')
            trainer.logger.raw(meter)
        else:
            trainer.logger.info(f'Loaded.')

    def on_imodels_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        trainer.logger.info('Model initialized.')

    def on_prepare_dataloader_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        res = self.getfuncargs(func, *args, **kwargs)
        stage = res['stage'].name
        loader_ = trainer.datamodule[stage]
        if loader_ is None:
            trainer.logger.info(f'{stage.capitalize()} dataloader prepared but no dataloader created.')
        else:
            trainer.logger.info(f'{stage.capitalize()} dataloader prepared, size: {len(loader_)}.')


class MeterCheckpoint(TrainCallback):
    """
    用于检视训练过程中模型的某个指标，并根据其提升进行 checkpoint 类型的保存
    该类参考了 Keras 中相应的实现。
    """
    only_main_process = True

    def __init__(self, monitor, mode="train", lower=True, start_epoch=0):
        self.monitor = monitor
        self.mode = mode
        self.lower = lower
        self.last_val = NoneItem()
        self.start_epoch = start_epoch

    def on_train_epoch_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        meter = result.meter
        self.update("train", trainer, params, meter)

    def update(self, cur_mode, trainer, param, meter):
        if cur_mode != self.mode:
            return
        if param.eidx > self.start_epoch:
            item = meter[self.monitor]
            if isinstance(item, AvgItem):
                item = item.avg
            if isinstance(item, NoneItem):
                return

            if self.lower:
                if self.last_val > item:
                    trainer.logger.info("model imporved from {} to {}".format(self.last_val, item))
                    trainer.save_checkpoint(meter.serialize())
                    self.last_val = item
            else:
                if self.last_val < item:
                    trainer.logger.info("model imporved from {} to {}".format(self.last_val, item))
                    trainer.save_checkpoint(meter.serialize())
                    self.last_val = item

    def on_test_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        meter = result.meter
        self.update("test", trainer, params, meter)

    def on_eval_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        meter = result.meter
        self.update("eval", trainer, params, meter)

    def __repr__(self) -> str:
        return self._repr_by_val("monitor", "mode", "lower", "start_epoch")


class EpochCheckpoint(TrainCallback):
    """
    在 Trainer 训练过程中定时保存模型
    """
    only_main_process = True

    def __init__(self, per_epoch=50):
        self.per_epoch = per_epoch

    def on_train_epoch_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        meter = result.meter
        if params.eidx % self.per_epoch == 0 and params.eidx > 0:
            trainer.save_keypoint(meta_info=trainer._wrap_result(meter))

    def __repr__(self) -> str:
        return self._repr_by_val("per_epoch")


class GlobalStepCheckpoint(TrainCallback):
    only_main_process = True

    def __init__(self, per_step=2500):
        self.per = per_step

    def on_train_step_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        super().on_train_step_end(trainer, func, params, meter, *args, **kwargs)
        if params.global_step % self.per == 0 and params.global_step > 0:
            trainer.save_checkpoint(meta_info=trainer._wrap_result(meter))


class KeyErrorSave(TrainCallback):
    only_main_process = True
    only_single_gpu = True
    priority = -1

    def __init__(self, wait_input=False):
        self.wait_input = wait_input

    def on_first_exception(self, source: Trainer, func, params: Params, e: BaseException, *args, **kwargs):
        if isinstance(e, (KeyboardInterrupt)):
            source.logger.info("KeyErrorSave trigged, save checkpoint")
            source.save_keypoint({"mode": "KeyboardInterrupt"})

            tp = "n"
            if self.wait_input:
                tp = input("continue train step? (y/other)")

            if tp.lower() == "y":
                return True


class BoardRecord(TrainCallback):
    """
    自动记录训练过程中的所有变量到 tensorboard 中（epoch 级）
    """
    only_main_process = True
    priority = 100

    def __init__(self) -> None:
        super().__init__()

    def on_hooked(self, source: Trainer, params: Params):
        self.start = 0

    def _key_name(self, mode, key):
        return "{}_{}_".format(key, mode)

    def on_test_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        meter = result.meter
        if isinstance(meter, Meter):
            for k, v in meter.numeral_items():
                trainer.writer.add_scalar(self._key_name("test", k), v, params.eidx)

    def on_train_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        self.start = params.eidx

    def on_eval_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        meter = result.meter
        if isinstance(meter, Meter):
            for k, v in meter.numeral_items():
                trainer.writer.add_scalar(self._key_name("eval", k), v, params.eidx)

    def on_train_epoch_end(self, trainer: Trainer, func, params: Params, meter: AvgMeter, *args, **kwargs):
        if isinstance(meter, Meter):
            for k, v in meter.numeral_items():
                trainer.writer.add_scalar(self._key_name("train", k), v, params.eidx)


class EMAUpdate(TrainCallback):
    only_main_process = True

    def on_train_step_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        super().on_train_step_end(trainer, func, params, meter, *args, **kwargs)
        for k, v in trainer.model_dict.items():
            if k.lower().startswith('ema'):
                v.step()


class LRSchedule(TrainCallback):
    def __init__(self, schedule: Schedule = None, apply=True, use_eidx=True):
        self.schedule = schedule
        self.apply = apply
        self.use_eidx = use_eidx

    def on_hooked(self, source: Trainer, params: Params):
        super().on_hooked(source, params)
        if self.schedule is None:
            if 'lr_sche' not in params:
                source.logger.warn(f'lr_sche not exists in params and be assigned, {self} will be unhooked after.')
                self.unhook()
            else:
                self.schedule = params.lr_sche

    def on_train_epoch_begin(self, trainer: Trainer, func, params: Params, *args, **kwargs):
        for k, v in trainer.optim_dict.items():
            if self.use_eidx:
                step = params.eidx
            else:
                step = params.global_step

            if self.apply:
                new_lr = self.schedule.apply(v, step)
                trainer.logger.info('{}.lr = {}'.format(k, new_lr))
            else:
                ratio = self.schedule.scale(v, step)
                trainer.logger.info('lr scale ratio = {}'.format(k, ratio))


class ReportSche(TrainCallback):
    """
    log `schedule` in every epoch end
    `schedule` means `Schedule` in Params and have `sche` in the name, which will have different value in every epoch
    """
    only_main_process = True
    priority = 100

    def on_hooked(self, source: Trainer, params: Params):
        self.sche_lis = []
        for k, v in params.items():  # type:str, Any
            if isinstance(v, (Schedule, ScheduleList)) and 'sche' in k.lower():
                self.sche_lis.append((k, v))

    def on_train_epoch_end(self, trainer: Trainer, func, params: Params, result: TrainerResult, *args, **kwargs):
        super().on_train_epoch_end(trainer, func, params, result, *args, **kwargs)
        meter = result.meter
        m = Meter()
        for k, v in self.sche_lis:
            m[k] = v(params.eidx)
        trainer.logger.info(m)


class AutoLoadModel(InitialCallback):

    def on_imodels_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        super().on_imodels_end(trainer, func, params, meter, *args, **kwargs)
        if params.get('pretrain', False):
            path = params.get('pretrain_path', None)
            if path is not None:
                trainer.load_state_dict(path)


class EvalFirst(AutoLoadModel):

    def __init__(self, datamodule=None):
        super().__init__()
        self.datamodule = datamodule

    def on_imodels_end(self, trainer: Trainer, func, params: Params, meter: Meter, *args, **kwargs):
        super().on_imodels_end(trainer, func, params, meter, *args, **kwargs)
        if params.get('eval_first', True):
            if self.datamodule is not None:
                trainer.evaluate(self.datamodule)
            else:
                trainer.evaluate()



import sys
import textwrap
import traceback


def safe_load(default=None, print_exc_stack=True, print_method_args=True):
    def outer(func):
        def inner(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except BaseException as e:
                traceback.print_exc()
                print('=' * 80, file=sys.stderr)
                size = 80
                print('| ignored exception from safe_load()', file=sys.stderr)
                if print_method_args:
                    arg_info = f"| {args, kwargs}"
                    print(arg_info, file=sys.stderr)
                    size = len(arg_info)
                if print_exc_stack:
                    exc_info = traceback.format_exc().strip()
                    size = max(len(i) for i in exc_info.split('\n')) + 10
                    print(textwrap.indent(exc_info, '| '), file=sys.stderr)
                print('=' * size, file=sys.stderr)
                return default

        return inner

    return outer


def safe_dump(print_exc_stack=True, print_method_args=True):
    def outer(func):
        def inner(*args, **kwargs):
            try:
                func(*args, **kwargs)
                return True
            except BaseException as e:
                print('=' * 20, file=sys.stderr)
                if print_method_args:
                    print(args, kwargs, file=sys.stderr)
                if print_exc_stack:
                    traceback.print_exc()
                print('=' * 20, file=sys.stderr)
                return False

        return inner

    return outer



# copied from https://github.com/sksq96/pytorch-summary

import torch
import torch.nn as nn
from torch.autograd import Variable

from collections import OrderedDict
import numpy as np


def summary(model, input_size, batch_size=-1, device=torch.device('cuda:0'), dtypes=None):
    result, params_info = summary_string(
        model, input_size, batch_size, device, dtypes)
    print(result)

    return params_info


def summary_string(model, input_size, batch_size=-1, device=torch.device('cuda:0'), dtypes=None):
    if dtypes == None:
        dtypes = [torch.FloatTensor]*len(input_size)

    summary_str = ''

    def register_hook(module):
        def hook(module, input, output):
            class_name = str(module.__class__).split(".")[-1].split("'")[0]
            module_idx = len(summary)

            m_key = "%s-%i" % (class_name, module_idx + 1)
            summary[m_key] = OrderedDict()
            summary[m_key]["input_shape"] = list(input[0].size())
            summary[m_key]["input_shape"][0] = batch_size
            if isinstance(output, (list, tuple)):
                summary[m_key]["output_shape"] = [
                    [-1] + list(o.size())[1:] for o in output
                ]
            else:
                summary[m_key]["output_shape"] = list(output.size())
                summary[m_key]["output_shape"][0] = batch_size

            params = 0
            if hasattr(module, "weight") and hasattr(module.weight, "size"):
                params += torch.prod(torch.LongTensor(list(module.weight.size())))
                summary[m_key]["trainable"] = module.weight.requires_grad
            if hasattr(module, "bias") and hasattr(module.bias, "size"):
                params += torch.prod(torch.LongTensor(list(module.bias.size())))
            summary[m_key]["nb_params"] = params

        if (
                not isinstance(module, nn.Sequential)
                and not isinstance(module, nn.ModuleList)
        ):
            hooks.append(module.register_forward_hook(hook))

    # multiple inputs to the network
    if isinstance(input_size, tuple):
        input_size = [input_size]

    # batch_size of 2 for batchnorm
    x = [torch.rand(2, *in_size).type(dtype).to(device=device)
         for in_size, dtype in zip(input_size, dtypes)]

    # create properties
    summary = OrderedDict()
    hooks = []

    # register hook
    model.apply(register_hook)

    # make a forward pass
    # print(x.shape)
    model(*x)

    # remove these hooks
    for h in hooks:
        h.remove()

    summary_str += "----------------------------------------------------------------" + "\n"
    line_new = "{:>20}  {:>25} {:>15}".format(
        "Layer (type)", "Output Shape", "Param #")
    summary_str += line_new + "\n"
    summary_str += "================================================================" + "\n"
    total_params = 0
    total_output = 0
    trainable_params = 0
    for layer in summary:
        # input_shape, output_shape, trainable, nb_params
        line_new = "{:>20}  {:>25} {:>15}".format(
            layer,
            str(summary[layer]["output_shape"]),
            "{0:,}".format(summary[layer]["nb_params"]),
        )
        total_params += summary[layer]["nb_params"]

        total_output += np.prod(summary[layer]["output_shape"])
        if "trainable" in summary[layer]:
            if summary[layer]["trainable"] == True:
                trainable_params += summary[layer]["nb_params"]
        summary_str += line_new + "\n"

    # assume 4 bytes/number (float on cuda).
    total_input_size = abs(np.prod(sum(input_size, ()))
                           * batch_size * 4. / (1024 ** 2.))
    total_output_size = abs(2. * total_output * 4. /
                            (1024 ** 2.))  # x2 for gradients
    total_params_size = abs(total_params * 4. / (1024 ** 2.))
    total_size = total_params_size + total_output_size + total_input_size

    summary_str += "================================================================" + "\n"
    summary_str += "Total params: {0:,}".format(total_params) + "\n"
    summary_str += "Trainable params: {0:,}".format(trainable_params) + "\n"
    summary_str += "Non-trainable params: {0:,}".format(total_params -
                                                        trainable_params) + "\n"
    summary_str += "----------------------------------------------------------------" + "\n"
    summary_str += "Input size (MB): %0.2f" % total_input_size + "\n"
    summary_str += "Forward/backward pass size (MB): %0.2f" % total_output_size + "\n"
    summary_str += "Params size (MB): %0.2f" % total_params_size + "\n"
    summary_str += "Estimated Total Size (MB): %0.2f" % total_size + "\n"
    summary_str += "----------------------------------------------------------------" + "\n"
    # return summary
    return summary_str, (total_params, trainable_params)


from typing import Union, Dict, List, Callable
from functools import partial
from collections import OrderedDict


def regist_func_to(val: Union[Dict[str, Callable], List[Callable]], name_=None):
    def wrap(func):
        if name_ is None:
            name = func.__name__
        else:
            name = name_
        if isinstance(val, dict):
            val[name] = func
        elif isinstance(val, list):
            val.append(func)

        return func

    return wrap


class Register():
    def __init__(self, name):
        self.name = name
        self.source = OrderedDict()

    def __str__(self):
        inner = str([(k, v) for k, v in self.source.items()])
        return f"Register({self.name}{inner})"

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, item):
        return self.source.get(item, None)

    def __call__(self, wrapped, name=None):
        if name is None:
            name = wrapped.__name__
        assert name is not None
        self.source[name] = wrapped

    def regist(self, name=None):
        return partial(self, name=name)



import os
from typing import Tuple, Iterable, List, Union


def walk(top: str, depth: int = -1) -> Iterable[Tuple[str, List[str], List[str]]]:
    """same as os.path.walk but added depth argument"""
    fs = os.listdir(top)

    dirs = []
    files = []
    rfs = [os.path.join(top, i) for i in fs]
    [dirs.append(i) if os.path.isdir(j) else files.append(i) for i, j in zip(fs, rfs)]
    yield top, dirs, files
    if depth > 0:
        for dir in dirs:
            yield from walk(os.path.join(top, dir), depth - 1)


def walk_file(root, suffix_lis: Union[List[str], set]):
    suffix_lis = {f".{i.lstrip('.')}" for i in suffix_lis}
    for root, dirs, fs in os.walk(root):
        yield from [os.path.join(root, f) for f in fs if os.path.splitext(f)[1] in suffix_lis]



from typing import Sequence, Union, Dict, Any
import os
import shutil

import numpy as np
import torch
from torch.utils.data import Dataset


class Data:
    def __init__(self, value, name=None):
        self.name = name
        self.value = value


class DelegateDataTypeError(RuntimeError): pass


class DataDelegate(Dataset):
    """
    """

    def __init__(self, source, len=None, len_func=None):
        self.source = source
        self.len = len
        self.len_func = len_func

    def __len__(self):
        if self.len_func is not None:
            return self.len_func(self.source)
        if self.len is not None:
            return self.len
        return len(self.source)

    def __getitem__(self, item) -> Union[Data, Sequence[Data], Dict[str, Data], Dict[str, Any]]:
        raise NotImplementedError()

    def create_data(self, value, name=None) -> Data:
        return Data(value=value, name=name)


class LambdaDelegate(DataDelegate):
    """
    It's not recommaned to use lambda function to feed `get_func` or `len_func`, it will raise an error
    in MultiProcess mode cause <lambda> function is unpicklable object.
    """

    def __init__(self, name, get_func, source, len=None, len_func=None, wrap=True):
        super().__init__(source, len, len_func)
        self.name = name
        self.get_func = get_func
        self.wrap = wrap

    def __getitem__(self, index):
        val = self.get_func(index)
        if self.wrap:
            return Data(val, self.name)
        return val


class ClassSampleDelegate(DataDelegate):

    def __init__(self, source, len=None, len_func=None):
        super().__init__(source, len, len_func)
        self._n_classes = list()
        self._neg_classes = dict()
        self._cls_dict = {}

    def _create_cls_dict(self, cls_squence: Sequence[int]):
        if isinstance(cls_squence, torch.Tensor):
            cls_squence = cls_squence.reshape(-1).detach().cpu().numpy()
        elif isinstance(cls_squence, np.ndarray):
            cls_squence = cls_squence.view(-1)
        elif isinstance(cls_squence, list):
            cls_squence = np.array(cls_squence)
        else:
            assert False

        self._n_classes = list(set(cls_squence))
        neg_cls = {i: list(self._n_classes) for i in self._n_classes}
        _ = {neg_cls[i].remove(i) for i in neg_cls}
        self._neg_classes = neg_cls

        for i in self._n_classes:
            self._cls_dict[i] = np.where(cls_squence == i)[0]

    def _rand_choice_same_cls(self, cls):
        return np.random.choice(self._cls_dict[cls], 1)

    def _rand_choice_neg_cls(self, cls):
        neg_cls = np.random.choice(self._neg_classes[cls], 1)
        return np.random.choice(self._cls_dict[neg_cls], 1)


class BaseHDF5Builder(DataDelegate):
    def __init__(self, h5path: str):
        super().__init__(None)
        import h5py
        self.h5path = h5path
        try:
            self.h5file = h5py.File(h5path, 'r')
        except OSError:
            self._create_tmp_h5(h5path)
        self.initial_h5attr()

    def __len__(self):
        raise NotImplementedError()

    def __getitem__(self, index):
        raise NotImplementedError()

    def _create_tmp_h5(self, h5path):
        import h5py
        c = 0
        h5file = None
        while h5file is not None:
            nh5path = f"{h5path}_builder_tmp{c}.h5"
            try:
                if not os.path.exists(nh5path):
                    shutil.copy(h5path, nh5path)
                h5file = h5py.File(nh5path, 'r')
            except OSError:
                c += 1
        return h5file

    def initial_h5attr(self):
        pass



def dicmap(kout=None, **maps):
    def wrap(function):
        def inner(mem: dict):
            fkwargs = {k: mem[v] for k, v in maps.items()}
            res = function(**fkwargs)
            if isinstance(res, dict):
                for k in res:
                    mem[k] = res[k]
            elif isinstance(res, (list, tuple)) and isinstance(kout, (list, tuple)):
                assert len(res) == len(kout)
                for k, v in zip(kout, res):
                    mem[k] = v
            elif isinstance(kout, str):
                mem[kout] = res

        return inner

    return wrap



"""
Help you find the experiments you have done.
"""

import os
import shutil
from collections import OrderedDict
from dataclasses import dataclass
from functools import lru_cache
from typing import Dict, List

from git import Commit, Repo
from lumo.base_classes import attr
from lumo.utils import safe_io as io
from lumo.proc.date import date_from_str
from lumo.proc.const import FN, EXP
from lumo.utils.paths import compare_path
from lumo.proc.path import git_dir, libhome


class TestProp():
    @property
    def command(self):
        raise NotImplementedError()

    @property
    def exp_name(self):
        raise NotImplementedError()

    @property
    def exp_root(self):
        raise NotImplementedError()

    @property
    def project_name(self):
        raise NotImplementedError()

    @property
    def repo_root(self):
        raise NotImplementedError()

    @property
    def repo_hash(self):
        raise NotImplementedError()

    @property
    def start_time(self):
        raise NotImplementedError()

    @property
    def end_time(self):
        raise NotImplementedError()

    @property
    def end_code(self):
        raise NotImplementedError()

    @property
    def grepo(self) -> Repo:
        raise NotImplementedError()

    @property
    def gcommit(self) -> Commit:
        raise NotImplementedError()

    @property
    def commit_hash(self) -> str:
        raise NotImplementedError()

    @property
    @lru_cache(1)
    def lines(self) -> Dict[str, Dict]:
        raise NotImplementedError()

    @property
    @lru_cache(1)
    def jsons(self):
        raise NotImplementedError()

    @property
    def pkl_keys(self):
        raise NotImplementedError()

    @property
    def uuid(self):
        raise NotImplementedError()

    @property
    def largest_epoch(self):
        raise NotImplementedError()


@dataclass(unsafe_hash=True)
class Test(TestProp):
    tid: str
    root: str

    @property
    def command(self):
        exe = self.jsons[EXP.EXECUTE]['exec_bin']
        argv = self.jsons[EXP.EXECUTE]['exec_argv']
        return ' '.join([exe, *argv])

    @property
    def exp_name(self):
        return os.path.basename(self.exp_root)

    @property
    def exp_root(self):
        return os.path.dirname(self.root)

    @property
    def project_name(self):
        return os.path.basename(self.repo_root)

    @property
    def repo_root(self):
        return self.jsons[EXP.PROJECT]['root']

    @property
    def repo_hash(self):
        return self.jsons[EXP.PROJECT]['hash']

    @property
    def start_time(self):
        val = self.jsons[EXP.STATE].get('start', None)
        if val is not None:
            val = date_from_str(val)
        return val

    @property
    def end_time(self):
        val = self.jsons[EXP.STATE].get('end', None)
        if val is not None:
            val = date_from_str(val)
        return val

    @property
    def end_code(self):
        return self.jsons[EXP.STATE].get('end_code', 1)

    @property
    def grepo(self) -> Repo:
        raise NotImplementedError()
        # repo = load_repo(self.repo_root)
        # return repo

    @property
    def gcommit(self) -> Commit:
        from gitdb.util import hex_to_bin
        return Commit(self.grepo, hex_to_bin(self.commit_hash))

    @property
    def commit_hash(self) -> str:
        """None if nocommit in this test, or commit hash will be returned"""
        return self.jsons.get(EXP.GIT, {}).get('commit', None)

    @property
    def lines(self) -> Dict[str, str]:
        dir_ = os.path.join(self.root, FN.D_LINE)
        if not os.path.exists(dir_):
            return attr()
        fs_ = os.listdir(dir_)
        fs_ = [i for i in fs_ if i.endswith(FN.SUFFIX.D_LINE)]
        res = attr()
        for f in fs_:
            k = os.path.splitext(f)[0]
            v = io.load_string(os.path.join(dir_, f))
            if v is None:
                os.remove(os.path.join(dir_, f))
                continue
            res[k] = v
        return res

    @property
    def jsons(self):
        dir_ = os.path.join(self.root, FN.D_JSON)
        if not os.path.exists(dir_):
            return attr()
        fs_ = os.listdir(dir_)
        fs_ = [i for i in fs_ if i.endswith('.json')]
        res = attr()
        for f in fs_:
            k = os.path.splitext(f)[0]
            v = io.load_json(os.path.join(dir_, f))
            if v is None:
                continue
            res[k] = v
        return res

    @property
    @lru_cache(1)
    def pkl_keys(self):
        dir_ = os.path.join(self.root, FN.D_JSON)
        if not os.path.exists(dir_):
            return set()

        fs_ = os.listdir(dir_)
        fs_ = [i for i in fs_ if i.endswith('.pkl')]
        res = set()
        for f in fs_:
            k = os.path.splitext(f)[0]
            res.add(k)
        return res

    @property
    def uuid(self):
        return self.lines.get(EXP.UUID, '')

    @property
    def largest_epoch(self):
        return self.jsons.get(EXP.TRAINER, {}).get('epoch', 0)

    def pkl(self, key):
        """

        Args:
            key:

        Returns:

        """
        dir_ = os.path.join(self.root, FN.D_JSON)
        f = os.path.join(dir_, f"{key}.pkl")
        return io.load_state_dict(f)


class Condition():
    def filter(self, test: Test) -> bool:
        raise NotImplementedError()

    def __call__(self, tests: List[Test]):
        return [i for i in tests if self.filter(i)]


class LambdaCondition(Condition):
    def __init__(self, func):
        self.func = func

    def filter(self, test: Test) -> bool:
        return self.func(test)


class Identity(Condition):

    def filter(self, test: Test) -> bool:
        return True


class Or(Condition):
    def __init__(self, *conditions: Condition):
        self.conditions = list(conditions)

    def add(self, condition: Condition):
        self.conditions.append(condition)

    def filter(self, test: Test) -> bool:
        return any([i.filter(test) for i in self.conditions])


class And(Or):
    def filter(self, test: Test) -> bool:
        return all([i.filter(test) for i in self.conditions])


class Query:
    def __init__(self):
        self.conditions = And()

    def success(self):
        self.conditions.add(LambdaCondition(lambda x: x.end_code == 0))
        return self

    def failed(self):
        self.conditions.add(LambdaCondition(lambda x: x.end_code != 0))
        return self

    def has_info(self, key):
        self.conditions.add(LambdaCondition(lambda x: x.jsons.get(key, None) is not None))
        return self

    def has_line(self, key):
        self.conditions.add(LambdaCondition(lambda x: x.lines.get(key, None) is not None))
        return self

    def in_time(self, start=None, end=None):
        if start is not None:
            self.conditions.add(LambdaCondition(lambda x: x.start_time > start is not None))
        if end is not None:
            self.conditions.add(LambdaCondition(lambda x: x.end_time > end is not None))
        return self

    def train_longer_than(self, epoch=0):
        self.conditions.add(LambdaCondition(lambda x: x.largest_epoch > epoch))
        return self

    def train_shorter_than(self, epoch=1):
        self.conditions.add(LambdaCondition(lambda x: x.largest_epoch <= epoch))
        return self

    def in_repo(self, name=None, path=None, full=False):
        if name is not None:
            if full:
                self.conditions.add(LambdaCondition(lambda x: name == x.project_name))  # type:Test
            else:
                self.conditions.add(LambdaCondition(lambda x: name in x.project_name))  # type:Test
        elif path is not None:
            self.conditions.add(LambdaCondition(lambda x: compare_path(path, x.repo_root)))  # type:Test
        else:
            path = git_dir(path)
            if path is not None:
                return self.in_repo(path=path)
            else:
                return Identity()

        return self

    def build(self):
        return self.conditions


class Finder:
    def __init__(self):
        fn = os.path.join(libhome(), FN.REPOSJS)
        if os.path.exists(fn):
            res = io.load_json(fn)
        else:
            res = {}
        self.meta = res
        self._test_dirs = []
        self._tests = Tests()
        self.refresh()

    def refresh(self):
        fn = os.path.join(libhome(), FN.TESTLOG)
        if not os.path.exists(fn):
            return
        res = io.load_string(fn)
        if res is None:
            res = ''
        res = res.split('\n')

        self._test_dirs = list(OrderedDict.fromkeys(
            filter(lambda x: os.path.exists(x.strip()), res)
        ).keys())
        self._test_dirs = sorted(self._test_dirs, key=lambda x: os.stat(x).st_atime)

        self._tests = Tests(Test(os.path.basename(i), i) for i in self._test_dirs)

    def tests(self, *conditions, nonflag=None):
        """can be found in each expdir"""
        tests = self._tests
        for condition in conditions:
            tests = condition(tests)
        if len(tests) == 0:
            return nonflag
        return Tests(tests)

    def lastest(self) -> Test:
        return self._tests[-1]

    def n_lastest(self, n) -> List[Test]:
        return self._tests[-n:]


class Tests(TestProp, List[Test]):

    def _to_df(self, value: list, column: str):
        import pandas as pd
        df = pd.DataFrame()
        df[column] = value
        df.index = self.names
        return df

    @property
    def command(self):
        return self._to_df([i.command for i in self], 'command')

    @property
    def exp_name(self):
        return self._to_df([i.exp_name for i in self], 'exp_name')

    @property
    def exp_root(self):
        return self._to_df([i.exp_root for i in self], 'exp_root')

    @property
    def project_name(self):
        return self._to_df([i.project_name for i in self], 'project_name')

    @property
    def repo_root(self):
        return self._to_df([i.repo_root for i in self], 'repo_root')

    @property
    def repo_hash(self):
        return self._to_df([i.repo_hash for i in self], 'repo_hash')

    @property
    def start_time(self):
        return self._to_df([i.start_time for i in self], 'start_time')

    @property
    def end_time(self):
        return self._to_df([i.end_time for i in self], 'end_time')

    @property
    def end_code(self):
        return self._to_df([i.end_code for i in self], 'end_code')

    @property
    def grepo(self):
        return self._to_df([i.grepo for i in self], 'grepo')

    @property
    def gcommit(self):
        return self._to_df([i.gcommit for i in self], 'gcommit')

    @property
    def commit_hash(self):
        return self._to_df([i.commit_hash for i in self], 'commit_hash')

    @property
    def lines(self):
        return [i.lines for i in self]

    @property
    def jsons(self):
        return [i.jsons for i in self]

    @property
    def pkl_keys(self):
        return [i.pkl_keys for i in self]

    @property
    def names(self):
        return [i.tid for i in self]

    @property
    def uuid(self):
        return self._to_df([i.uuid for i in self], 'uuid')

    @property
    def latest(self):
        return self[-1]

    @property
    def largest_epoch(self):
        return self._to_df([i.largest_epoch for i in self], 'largest_epoch')

    @property
    def root(self):
        return self._to_df([i.root for i in self], 'root')

    def delete(self):
        for i in self:
            shutil.rmtree(i.root)


F = Finder()
Q = Query()
